# Copyright 2018 Christoph Reiter <reiter.christoph@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import pytest

from gi import PyGIDeprecationWarning

GdkPixbuf = pytest.importorskip("gi.repository.GdkPixbuf")


def test_new_from_data():
    width = 600
    height = 32769
    pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, height)
    pixels = pixbuf.get_pixels()
    new_pixbuf = GdkPixbuf.Pixbuf.new_from_data(
        pixels,
        GdkPixbuf.Colorspace.RGB,
        True,
        8,
        pixbuf.get_width(),
        pixbuf.get_height(),
        pixbuf.get_rowstride(),
    )
    del pixbuf
    del pixels
    new_pixels = new_pixbuf.get_pixels()
    assert len(new_pixels) == width * height * 4


def test_new_from_data_deprecated_args():
    GdkPixbuf.Pixbuf.new_from_data(b"1234", 0, True, 8, 1, 1, 4)
    GdkPixbuf.Pixbuf.new_from_data(b"1234", 0, True, 8, 1, 1, 4, None)
    with (
        pytest.warns(PyGIDeprecationWarning, match=".*destroy_fn argument.*"),
        pytest.warns(PyGIDeprecationWarning, match=".*destroy_fn_data argument.*"),
    ):
        GdkPixbuf.Pixbuf.new_from_data(
            b"1234", 0, True, 8, 1, 1, 4, object(), object(), object()
        )



import contextlib
import unittest
import traceback
import ctypes
import warnings
import sys
import os
import re
import platform
import gc
import timeit
import random

import pytest

from gi.repository import Regress as Everything
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gio

from .helper import capture_exceptions


const_str = b"const \xe2\x99\xa5 utf8".decode("UTF-8")
noconst_str = "non" + const_str


class RawGList(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("next", ctypes.c_void_p),
        ("prev", ctypes.c_void_p),
    ]

    @classmethod
    def from_wrapped(cls, obj):
        offset = sys.getsizeof(object())  # size of PyObject_HEAD
        return ctypes.POINTER(cls).from_address(id(obj) + offset)


class TestInstanceTransfer(unittest.TestCase):
    def test_main(self):
        obj = Everything.TestObj()
        for _ in range(10):
            obj.instance_method_full()


class TestEverything(unittest.TestCase):
    def test_bool(self):
        self.assertEqual(Everything.test_boolean(False), False)
        self.assertEqual(Everything.test_boolean(True), True)
        self.assertEqual(Everything.test_boolean("hello"), True)
        self.assertEqual(Everything.test_boolean(""), False)

        self.assertEqual(Everything.test_boolean_true(True), True)
        self.assertEqual(Everything.test_boolean_false(False), False)

    def test_int8(self):
        self.assertEqual(Everything.test_int8(GLib.MAXINT8), GLib.MAXINT8)
        self.assertEqual(Everything.test_int8(GLib.MININT8), GLib.MININT8)
        self.assertRaises(OverflowError, Everything.test_int8, GLib.MAXINT8 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXINT8 + 1} not in range {GLib.MININT8} to {GLib.MAXINT8}",
        ):
            Everything.test_int8(GLib.MAXINT8 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MININT8} to {GLib.MAXINT8}",
        ):
            Everything.test_int8(GLib.MAXUINT64 * 2)

    def test_uint8(self):
        self.assertEqual(Everything.test_uint8(GLib.MAXUINT8), GLib.MAXUINT8)
        self.assertEqual(Everything.test_uint8(0), 0)
        self.assertRaises(OverflowError, Everything.test_uint8, -1)
        self.assertRaises(OverflowError, Everything.test_uint8, GLib.MAXUINT8 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT8 + 1} not in range 0 to {GLib.MAXUINT8}",
        ):
            Everything.test_uint8(GLib.MAXUINT8 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUINT8}",
        ):
            Everything.test_uint8(GLib.MAXUINT64 * 2)

    def test_int16(self):
        self.assertEqual(Everything.test_int16(GLib.MAXINT16), GLib.MAXINT16)
        self.assertEqual(Everything.test_int16(GLib.MININT16), GLib.MININT16)

        with pytest.raises(OverflowError, match="32768 not in range -32768 to 32767"):
            Everything.test_int16(GLib.MAXINT16 + 1)

        with pytest.raises(
            OverflowError, match="36893488147419103230 not in range -32768 to 32767"
        ):
            Everything.test_int16(GLib.MAXUINT64 * 2)

    def test_uint16(self):
        self.assertEqual(Everything.test_uint16(GLib.MAXUINT16), GLib.MAXUINT16)
        self.assertEqual(Everything.test_uint16(0), 0)
        self.assertRaises(OverflowError, Everything.test_uint16, -1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT16 + 1} not in range 0 to {GLib.MAXUINT16}",
        ):
            Everything.test_uint16(GLib.MAXUINT16 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUINT16}",
        ):
            Everything.test_uint16(GLib.MAXUINT64 * 2)

    def test_int32(self):
        self.assertEqual(Everything.test_int32(GLib.MAXINT32), GLib.MAXINT32)
        self.assertEqual(Everything.test_int32(GLib.MININT32), GLib.MININT32)

        with pytest.raises(
            OverflowError, match="2147483648 not in range -2147483648 to 2147483647"
        ):
            Everything.test_int32(GLib.MAXINT32 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXINT64 + 1} not in range -2147483648 to 2147483647",
        ):
            Everything.test_int32(GLib.MAXINT64 + 1)

    def test_uint32(self):
        self.assertEqual(Everything.test_uint32(GLib.MAXUINT32), GLib.MAXUINT32)
        self.assertEqual(Everything.test_uint32(0), 0)
        self.assertRaises(OverflowError, Everything.test_uint32, -1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT32 + 1} not in range 0 to {GLib.MAXUINT32}",
        ):
            Everything.test_uint32(GLib.MAXUINT32 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUINT32}",
        ):
            Everything.test_uint32(GLib.MAXUINT64 * 2)

    def test_int64(self):
        self.assertEqual(Everything.test_int64(GLib.MAXINT64), GLib.MAXINT64)
        self.assertEqual(Everything.test_int64(GLib.MININT64), GLib.MININT64)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXINT64 + 1} not in range {GLib.MININT64} to {GLib.MAXINT64}",
        ):
            Everything.test_int64(GLib.MAXINT64 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MININT64} to {GLib.MAXINT64}",
        ):
            Everything.test_int64(GLib.MAXUINT64 * 2)

    def test_uint64(self):
        self.assertEqual(Everything.test_uint64(GLib.MAXUINT64), GLib.MAXUINT64)
        self.assertEqual(Everything.test_uint64(0), 0)
        self.assertRaises(OverflowError, Everything.test_uint64, -1)
        self.assertRaises(OverflowError, Everything.test_uint64, GLib.MAXUINT64 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 + 1} not in range 0 to {GLib.MAXUINT64}",
        ):
            Everything.test_uint64(GLib.MAXUINT64 + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUINT64}",
        ):
            Everything.test_uint64(GLib.MAXUINT64 * 2)

    def test_int(self):
        self.assertEqual(Everything.test_int(GLib.MAXINT), GLib.MAXINT)
        self.assertEqual(Everything.test_int(GLib.MININT), GLib.MININT)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXINT + 1} not in range {GLib.MININT} to {GLib.MAXINT}",
        ):
            Everything.test_int(GLib.MAXINT + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MININT} to {GLib.MAXINT}",
        ):
            Everything.test_int(GLib.MAXUINT64 * 2)

    def test_uint(self):
        self.assertEqual(Everything.test_uint(GLib.MAXUINT), GLib.MAXUINT)
        self.assertEqual(Everything.test_uint(0), 0)
        self.assertRaises(OverflowError, Everything.test_uint, -1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT + 1} not in range 0 to {GLib.MAXUINT}",
        ):
            Everything.test_uint(GLib.MAXUINT + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUINT}",
        ):
            Everything.test_uint(GLib.MAXUINT64 * 2)

    def test_short(self):
        self.assertEqual(Everything.test_short(GLib.MAXSHORT), GLib.MAXSHORT)
        self.assertEqual(Everything.test_short(GLib.MINSHORT), GLib.MINSHORT)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXSHORT + 1} not in range {GLib.MINSHORT} to {GLib.MAXSHORT}",
        ):
            Everything.test_short(GLib.MAXSHORT + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MINSHORT} to {GLib.MAXSHORT}",
        ):
            Everything.test_short(GLib.MAXUINT64 * 2)

    def test_ushort(self):
        self.assertEqual(Everything.test_ushort(GLib.MAXUSHORT), GLib.MAXUSHORT)
        self.assertEqual(Everything.test_ushort(0), 0)
        self.assertRaises(OverflowError, Everything.test_ushort, -1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUSHORT + 1} not in range 0 to {GLib.MAXUSHORT}",
        ):
            Everything.test_ushort(GLib.MAXUSHORT + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXUSHORT}",
        ):
            Everything.test_ushort(GLib.MAXUINT64 * 2)

    def test_long(self):
        self.assertEqual(Everything.test_long(GLib.MAXLONG), GLib.MAXLONG)
        self.assertEqual(Everything.test_long(GLib.MINLONG), GLib.MINLONG)
        self.assertRaises(OverflowError, Everything.test_long, GLib.MAXLONG + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXLONG + 1} not in range {GLib.MINLONG} to {GLib.MAXLONG}",
        ):
            Everything.test_long(GLib.MAXLONG + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MINLONG} to {GLib.MAXLONG}",
        ):
            Everything.test_long(GLib.MAXUINT64 * 2)

    def test_ulong(self):
        self.assertEqual(Everything.test_ulong(GLib.MAXULONG), GLib.MAXULONG)
        self.assertEqual(Everything.test_ulong(0), 0)
        self.assertRaises(OverflowError, Everything.test_ulong, -1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXULONG + 1} not in range 0 to {GLib.MAXULONG}",
        ):
            Everything.test_ulong(GLib.MAXULONG + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXULONG}",
        ):
            Everything.test_ulong(GLib.MAXUINT64 * 2)

    def test_ssize(self):
        self.assertEqual(Everything.test_ssize(GLib.MAXSSIZE), GLib.MAXSSIZE)
        self.assertEqual(Everything.test_ssize(GLib.MINSSIZE), GLib.MINSSIZE)
        self.assertRaises(OverflowError, Everything.test_ssize, GLib.MAXSSIZE + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXSSIZE + 1} not in range {GLib.MINSSIZE} to {GLib.MAXSSIZE}",
        ):
            Everything.test_ssize(GLib.MAXSSIZE + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range {GLib.MINSSIZE} to {GLib.MAXSSIZE}",
        ):
            Everything.test_ssize(GLib.MAXUINT64 * 2)

    def test_size(self):
        self.assertEqual(Everything.test_size(GLib.MAXSIZE), GLib.MAXSIZE)
        self.assertEqual(Everything.test_size(0), 0)
        self.assertRaises(OverflowError, Everything.test_size, -1)
        self.assertRaises(OverflowError, Everything.test_size, GLib.MAXSIZE + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXSIZE + 1} not in range 0 to {GLib.MAXSIZE}",
        ):
            Everything.test_size(GLib.MAXSIZE + 1)

        with pytest.raises(
            OverflowError,
            match=f"{GLib.MAXUINT64 * 2} not in range 0 to {GLib.MAXSIZE}",
        ):
            Everything.test_size(GLib.MAXUINT64 * 2)

    def test_timet(self):
        self.assertEqual(Everything.test_timet(42), 42)
        self.assertRaises(OverflowError, Everything.test_timet, GLib.MAXUINT64 + 1)

    def test_unichar(self):
        self.assertEqual("c", Everything.test_unichar("c"))
        self.assertEqual(
            chr(sys.maxunicode), Everything.test_unichar(chr(sys.maxunicode))
        )

        self.assertEqual("♥", Everything.test_unichar("♥"))
        self.assertRaises(TypeError, Everything.test_unichar, "")
        self.assertRaises(TypeError, Everything.test_unichar, "morethanonechar")

    def test_float(self):
        self.assertEqual(Everything.test_float(GLib.MAXFLOAT), GLib.MAXFLOAT)
        self.assertEqual(Everything.test_float(GLib.MINFLOAT), GLib.MINFLOAT)
        self.assertRaises(OverflowError, Everything.test_float, GLib.MAXFLOAT * 2)

        with pytest.raises(
            OverflowError,
            match=re.escape(
                f"{GLib.MAXFLOAT * 2} not in range {-GLib.MAXFLOAT} to {GLib.MAXFLOAT}"
            ),
        ):
            Everything.test_float(GLib.MAXFLOAT * 2)

    def test_double(self):
        self.assertEqual(Everything.test_double(GLib.MAXDOUBLE), GLib.MAXDOUBLE)
        self.assertEqual(Everything.test_double(GLib.MINDOUBLE), GLib.MINDOUBLE)

        (two, three) = Everything.test_multi_double_args(2.5)
        self.assertAlmostEqual(two, 5.0)
        self.assertAlmostEqual(three, 7.5)

    def test_value(self):
        self.assertEqual(Everything.test_int_value_arg(GLib.MAXINT), GLib.MAXINT)
        self.assertEqual(Everything.test_value_return(GLib.MAXINT), GLib.MAXINT)

    def test_variant(self):
        v = Everything.test_gvariant_i()
        self.assertEqual(v.get_type_string(), "i")
        self.assertEqual(v.get_int32(), 1)

        v = Everything.test_gvariant_s()
        self.assertEqual(v.get_type_string(), "s")
        self.assertEqual(v.get_string(), "one")

        v = Everything.test_gvariant_v()
        self.assertEqual(v.get_type_string(), "v")
        vi = v.get_variant()
        self.assertEqual(vi.get_type_string(), "s")
        self.assertEqual(vi.get_string(), "contents")

        v = Everything.test_gvariant_as()
        self.assertEqual(v.get_type_string(), "as")
        self.assertEqual(v.get_strv(), ["one", "two", "three"])

        v = Everything.test_gvariant_asv()
        self.assertEqual(v.get_type_string(), "a{sv}")
        self.assertEqual(v.lookup_value("nosuchkey", None), None)
        name = v.lookup_value("name", None)
        self.assertEqual(name.get_string(), "foo")
        timeout = v.lookup_value("timeout", None)
        self.assertEqual(timeout.get_int32(), 10)

    def test_utf8_const_return(self):
        self.assertEqual(Everything.test_utf8_const_return(), const_str)

    def test_utf8_nonconst_return(self):
        self.assertEqual(Everything.test_utf8_nonconst_return(), noconst_str)

    def test_utf8_out(self):
        self.assertEqual(Everything.test_utf8_out(), noconst_str)

    def test_utf8_const_in(self):
        Everything.test_utf8_const_in(const_str)

    def test_utf8_inout(self):
        self.assertEqual(Everything.test_utf8_inout(const_str), noconst_str)

    def test_filename_return(self):
        if os.name != "nt":
            result = [os.fsdecode(b"\xc3\xa5\xc3\xa4\xc3\xb6"), "/etc/fstab"]
        else:
            result = ["åäö", "/etc/fstab"]
        self.assertEqual(Everything.test_filename_return(), result)

    def test_int_out_utf8(self):
        # returns g_utf8_strlen() in out argument
        self.assertEqual(Everything.test_int_out_utf8(""), 0)
        self.assertEqual(Everything.test_int_out_utf8("hello world"), 11)
        self.assertEqual(Everything.test_int_out_utf8("åäö"), 3)

    def test_utf8_out_out(self):
        self.assertEqual(Everything.test_utf8_out_out(), ("first", "second"))

    def test_utf8_out_nonconst_return(self):
        self.assertEqual(
            Everything.test_utf8_out_nonconst_return(), ("first", "second")
        )

    def test_enum(self):
        self.assertEqual(
            Everything.test_enum_param(Everything.TestEnum.VALUE1), "value1"
        )
        self.assertEqual(
            Everything.test_enum_param(Everything.TestEnum.VALUE3), "value3"
        )
        self.assertRaises(TypeError, Everything.test_enum_param, "hello")

    @pytest.mark.xfail(
        "32bit" in platform.architecture() or platform.system() == "Windows",
        reason="Big enum value doesn't convert to 32 bit (signed) long",
    )
    def test_enum_unsigned(self):
        self.assertEqual(
            Everything.test_unsigned_enum_param(Everything.TestEnumUnsigned.VALUE1),
            "value1",
        )
        self.assertEqual(
            Everything.test_unsigned_enum_param(Everything.TestEnumUnsigned.VALUE2),
            "value2",
        )
        self.assertRaises(TypeError, Everything.test_unsigned_enum_param, "hello")

    def test_flags(self):
        result = Everything.global_get_flags_out()
        # assert that it's not an int
        self.assertEqual(type(result), Everything.TestFlags)
        self.assertEqual(
            result, Everything.TestFlags.FLAG1 | Everything.TestFlags.FLAG3
        )

    def test_floating(self):
        e = Everything.TestFloating()
        self.assertEqual(e.__grefcount__, 1)

        e = GObject.new(Everything.TestFloating)
        self.assertEqual(e.__grefcount__, 1)

        e = Everything.TestFloating.new()
        self.assertEqual(e.__grefcount__, 1)

    def test_caller_allocates(self):
        struct_a = Everything.TestStructA()
        struct_a.some_int = 10
        struct_a.some_int8 = 21
        struct_a.some_double = 3.14
        struct_a.some_enum = Everything.TestEnum.VALUE3

        struct_a_clone = struct_a.clone()
        self.assertTrue(struct_a != struct_a_clone)
        self.assertEqual(struct_a.some_int, struct_a_clone.some_int)
        self.assertEqual(struct_a.some_int8, struct_a_clone.some_int8)
        self.assertEqual(struct_a.some_double, struct_a_clone.some_double)
        self.assertEqual(struct_a.some_enum, struct_a_clone.some_enum)

        struct_b = Everything.TestStructB()
        struct_b.some_int8 = 8
        struct_b.nested_a.some_int = 20
        struct_b.nested_a.some_int8 = 12
        struct_b.nested_a.some_double = 333.3333
        struct_b.nested_a.some_enum = Everything.TestEnum.VALUE2

        struct_b_clone = struct_b.clone()
        self.assertTrue(struct_b != struct_b_clone)
        self.assertEqual(struct_b.some_int8, struct_b_clone.some_int8)
        self.assertEqual(struct_b.nested_a.some_int, struct_b_clone.nested_a.some_int)
        self.assertEqual(struct_b.nested_a.some_int8, struct_b_clone.nested_a.some_int8)
        self.assertEqual(
            struct_b.nested_a.some_double, struct_b_clone.nested_a.some_double
        )
        self.assertEqual(struct_b.nested_a.some_enum, struct_b_clone.nested_a.some_enum)

        struct_a = Everything.test_struct_a_parse("ignored")
        self.assertEqual(struct_a.some_int, 23)

    def test_wrong_type_of_arguments(self):
        try:
            Everything.test_int8()
        except TypeError:
            (_e_type, e) = sys.exc_info()[:2]
            self.assertEqual(
                e.args, ("Regress.test_int8() takes exactly 1 argument (0 given)",)
            )

    def test_gtypes(self):
        gchararray_gtype = GObject.type_from_name("gchararray")
        gtype = Everything.test_gtype(str)
        self.assertEqual(gchararray_gtype, gtype)
        gtype = Everything.test_gtype("gchararray")
        self.assertEqual(gchararray_gtype, gtype)
        gobject_gtype = GObject.GObject.__gtype__
        gtype = Everything.test_gtype(GObject.GObject)
        self.assertEqual(gobject_gtype, gtype)
        gtype = Everything.test_gtype("GObject")
        self.assertEqual(gobject_gtype, gtype)
        self.assertRaises(TypeError, Everything.test_gtype, "invalidgtype")

        class NotARegisteredClass:
            pass

        self.assertRaises(TypeError, Everything.test_gtype, NotARegisteredClass)

        class ARegisteredClass(GObject.GObject):
            __gtype_name__ = "EverythingTestsARegisteredClass"

        gtype = Everything.test_gtype("EverythingTestsARegisteredClass")
        self.assertEqual(ARegisteredClass.__gtype__, gtype)
        gtype = Everything.test_gtype(ARegisteredClass)
        self.assertEqual(ARegisteredClass.__gtype__, gtype)
        self.assertRaises(TypeError, Everything.test_gtype, "ARegisteredClass")

    def test_dir(self):
        attr_list = dir(Everything)

        # test that typelib attributes are listed
        self.assertIn("TestStructA", attr_list)

        # test that instance members are listed
        self.assertIn("_namespace", attr_list)
        self.assertIn("_version", attr_list)

        # test that there are no duplicates returned
        self.assertEqual(len(attr_list), len(set(attr_list)))

    def test_array_int_in_empty(self):
        self.assertEqual(Everything.test_array_int_in([]), 0)

    def test_array_int_in(self):
        self.assertEqual(Everything.test_array_int_in([1, 5, -2]), 4)

    def test_array_int_out(self):
        self.assertEqual(Everything.test_array_int_out(), [0, 1, 2, 3, 4])

    def test_array_int_full_out(self):
        self.assertEqual(Everything.test_array_int_full_out(), [0, 1, 2, 3, 4])

    def test_array_int_none_out(self):
        self.assertEqual(Everything.test_array_int_none_out(), [1, 2, 3, 4, 5])

    def test_array_int_inout(self):
        self.assertEqual(Everything.test_array_int_inout([1, 5, 42, -8]), [6, 43, -7])

    def test_array_int_inout_empty(self):
        self.assertEqual(Everything.test_array_int_inout([]), [])

    def test_array_gint8_in(self):
        self.assertEqual(Everything.test_array_gint8_in(b"\x01\x03\x05"), 9)
        self.assertEqual(Everything.test_array_gint8_in([1, 3, 5, -50]), -41)

    def test_array_gint16_in(self):
        self.assertEqual(
            Everything.test_array_gint16_in([256, 257, -1000, 10000]), 9513
        )

    def test_array_gint32_in(self):
        self.assertEqual(Everything.test_array_gint32_in([30000, 1, -2]), 29999)

    def test_array_gint64_in(self):
        self.assertEqual(Everything.test_array_gint64_in([2**33, 2**34]), 2**33 + 2**34)

    def test_array_gtype_in(self):
        self.assertEqual(
            Everything.test_array_gtype_in(
                [GObject.TYPE_STRING, GObject.TYPE_UINT64, GObject.TYPE_VARIANT]
            ),
            "[gchararray,guint64,GVariant,]",
        )

    def test_array_fixed_size_int_in(self):
        # fixed length of 5
        self.assertEqual(Everything.test_array_fixed_size_int_in([1, 2, -10, 5, 3]), 1)

    def test_array_fixed_size_int_in_error(self):
        self.assertRaises(
            ValueError, Everything.test_array_fixed_size_int_in, [1, 2, 3, 4]
        )
        self.assertRaises(
            ValueError, Everything.test_array_fixed_size_int_in, [1, 2, 3, 4, 5, 6]
        )

    def test_array_fixed_size_int_out(self):
        self.assertEqual(Everything.test_array_fixed_size_int_out(), [0, 1, 2, 3, 4])

    def test_array_fixed_size_int_return(self):
        self.assertEqual(Everything.test_array_fixed_size_int_return(), [0, 1, 2, 3, 4])

    def test_array_of_non_utf8_strings(self):
        with pytest.raises(UnicodeDecodeError):
            Everything.test_array_of_non_utf8_strings()

    def test_array_callback(self):
        TestCallbacks.called = 0

        def callback(ints, ints_length, strings, strings_length):
            self.assertEqual(ints, [-1, 0, 1, 2])
            self.assertEqual(strings, ["one", "two", "three"])
            TestCallbacks.called += 1
            return TestCallbacks.called

        Everything.test_array_callback(callback)
        self.assertEqual(TestCallbacks.called, 2)

    def test_garray_container_return(self):
        # GPtrArray transfer container
        result = Everything.test_garray_container_return()
        self.assertEqual(result, ["regress"])
        result = None

    def test_garray_full_return(self):
        # GPtrArray transfer full
        result = Everything.test_garray_full_return()
        self.assertEqual(result, ["regress"])
        result = None

    def test_strv_out(self):
        self.assertEqual(
            Everything.test_strv_out(), ["thanks", "for", "all", "the", "fish"]
        )

    def test_strv_out_c(self):
        self.assertEqual(
            Everything.test_strv_out_c(), ["thanks", "for", "all", "the", "fish"]
        )

    def test_strv_out_container(self):
        self.assertEqual(Everything.test_strv_out_container(), ["1", "2", "3"])

    def test_strv_outarg(self):
        self.assertEqual(Everything.test_strv_outarg(), ["1", "2", "3"])

    def test_strv_in_gvalue(self):
        self.assertEqual(Everything.test_strv_in_gvalue(), ["one", "two", "three"])

    def test_strv_in(self):
        Everything.test_strv_in(["1", "2", "3"])

    def test_glist(self):
        self.assertEqual(Everything.test_glist_nothing_return(), ["1", "2", "3"])
        self.assertEqual(Everything.test_glist_nothing_return2(), ["1", "2", "3"])
        self.assertEqual(Everything.test_glist_container_return(), ["1", "2", "3"])
        self.assertEqual(Everything.test_glist_everything_return(), ["1", "2", "3"])

        Everything.test_glist_nothing_in(["1", "2", "3"])
        Everything.test_glist_nothing_in2(["1", "2", "3"])

    def test_glist_gtype(self):
        Everything.test_glist_gtype_container_in(
            [Everything.TestObj, Everything.TestSubObj]
        )

    def test_gslist(self):
        self.assertEqual(Everything.test_gslist_nothing_return(), ["1", "2", "3"])
        self.assertEqual(Everything.test_gslist_nothing_return2(), ["1", "2", "3"])
        self.assertEqual(Everything.test_gslist_container_return(), ["1", "2", "3"])
        self.assertEqual(Everything.test_gslist_everything_return(), ["1", "2", "3"])

        Everything.test_gslist_nothing_in(["1", "2", "3"])
        Everything.test_gslist_nothing_in2(["1", "2", "3"])

    def test_hash_return(self):
        expected = {"foo": "bar", "baz": "bat", "qux": "quux"}

        self.assertEqual(Everything.test_ghash_null_return(), None)
        self.assertEqual(Everything.test_ghash_nothing_return(), expected)
        self.assertEqual(Everything.test_ghash_nothing_return(), expected)
        self.assertEqual(Everything.test_ghash_container_return(), expected)
        self.assertEqual(Everything.test_ghash_everything_return(), expected)

        result = Everything.test_ghash_gvalue_return()
        self.assertEqual(result["integer"], 12)
        self.assertEqual(result["boolean"], True)
        self.assertEqual(result["string"], "some text")
        self.assertEqual(result["strings"], ["first", "second", "third"])
        self.assertEqual(
            result["flags"], Everything.TestFlags.FLAG1 | Everything.TestFlags.FLAG3
        )
        self.assertEqual(result["enum"], Everything.TestEnum.VALUE2)
        result = None

    # FIXME: CRITICAL **: Unsupported type ghash
    def disabled_test_hash_return_nested(self):
        self.assertEqual(Everything.test_ghash_nested_everything_return(), {})
        self.assertEqual(Everything.test_ghash_nested_everything_return2(), {})

    def test_hash_in(self):
        expected = {"foo": "bar", "baz": "bat", "qux": "quux"}

        Everything.test_ghash_nothing_in(expected)
        Everything.test_ghash_nothing_in2(expected)

    def test_hash_in_with_typed_strv(self):
        class GStrv(list):
            __gtype__ = GObject.TYPE_STRV

        data = {
            "integer": 12,
            "boolean": True,
            "string": "some text",
            "strings": GStrv(["first", "second", "third"]),
            "flags": Everything.TestFlags.FLAG1 | Everything.TestFlags.FLAG3,
            "enum": Everything.TestEnum.VALUE2,
        }
        Everything.test_ghash_gvalue_in(data)
        data = None

    def test_hash_in_with_gvalue_strv(self):
        data = {
            "integer": 12,
            "boolean": True,
            "string": "some text",
            "strings": GObject.Value(GObject.TYPE_STRV, ["first", "second", "third"]),
            "flags": Everything.TestFlags.FLAG1 | Everything.TestFlags.FLAG3,
            "enum": Everything.TestEnum.VALUE2,
        }
        Everything.test_ghash_gvalue_in(data)
        data = None

    @unittest.skipIf(platform.python_implementation() == "PyPy", "CPython only")
    def test_struct_gpointer(self):
        glist = GLib.List()
        raw = RawGList.from_wrapped(glist)

        # Note that pointer fields use 0 for NULL in PyGObject and None in ctypes
        self.assertEqual(glist.data, 0)
        self.assertEqual(raw.contents.data, None)

        glist.data = 123
        self.assertEqual(glist.data, 123)
        self.assertEqual(raw.contents.data, 123)

        glist.data = None
        self.assertEqual(glist.data, 0)
        self.assertEqual(raw.contents.data, None)

        # Setting to anything other than an int should raise
        self.assertRaises(TypeError, setattr, glist.data, "nan")
        self.assertRaises(TypeError, setattr, glist.data, object())
        self.assertRaises(TypeError, setattr, glist.data, 123.321)

    def test_struct_opaque(self):
        # we should get a sensible error message
        try:
            Everything.TestBoxedPrivate()
            self.fail(
                "allocating disguised struct without default constructor unexpectedly succeeded"
            )
        except TypeError:
            (e_type, e_value, e_tb) = sys.exc_info()
            self.assertEqual(e_type, TypeError)
            self.assertTrue("TestBoxedPrivate" in str(e_value), str(e_value))
            self.assertTrue("constructor" in str(e_value), str(e_value))
            tb = "".join(traceback.format_exception(e_type, e_value, e_tb))
            self.assertTrue('test_everything.py", line' in tb, tb)


class TestNullableArgs(unittest.TestCase):
    def test_in_nullable_hash(self):
        Everything.test_ghash_null_in(None)

    def test_in_nullable_list(self):
        Everything.test_gslist_null_in(None)
        Everything.test_glist_null_in(None)
        Everything.test_gslist_null_in([])
        Everything.test_glist_null_in([])

    def test_in_nullable_array(self):
        Everything.test_array_int_null_in(None)
        Everything.test_array_int_null_in([])

    def test_in_nullable_string(self):
        Everything.test_utf8_null_in(None)

    def test_in_nullable_object(self):
        Everything.func_obj_null_in(None)

    def test_out_nullable_hash(self):
        self.assertEqual(None, Everything.test_ghash_null_out())

    def test_out_nullable_list(self):
        self.assertEqual([], Everything.test_gslist_null_out())
        self.assertEqual([], Everything.test_glist_null_out())

    def test_out_nullable_array(self):
        self.assertEqual([], Everything.test_array_int_null_out())

    def test_out_nullable_string(self):
        self.assertEqual(None, Everything.test_utf8_null_out())

    def test_out_nullable_object(self):
        self.assertEqual(None, Everything.TestObj.null_out())


class TestCallbacks(unittest.TestCase):
    called = False
    main_loop = GLib.MainLoop()

    def test_callback(self):
        TestCallbacks.called = False

        def callback():
            TestCallbacks.called = True

        Everything.test_simple_callback(callback)
        self.assertTrue(TestCallbacks.called)

    def test_callback_exception(self):
        """This test ensures that we get errors from callbacks correctly
        and in particular that we do not segv when callbacks fail.
        """

        def callback():
            x = 1 / 0
            self.fail("unexpected surviving zero divsion:" + str(x))

        # note that we do NOT expect the ZeroDivisionError to be propagated
        # through from the callback, as it crosses the Python<->C boundary
        # twice. (See GNOME #616279)
        with capture_exceptions() as exc:
            Everything.test_simple_callback(callback)
        self.assertTrue(exc)
        self.assertEqual(exc[0].type, ZeroDivisionError)

    def test_double_callback_exception(self):
        """This test ensures that we get errors from callbacks correctly
        and in particular that we do not segv when callbacks fail.
        """

        def badcallback():
            x = 1 / 0
            self.fail("unexpected surviving zero divsion:" + str(x))

        def callback():
            Everything.test_boolean(True)
            Everything.test_boolean(False)
            Everything.test_simple_callback(badcallback())

        # note that we do NOT expect the ZeroDivisionError to be propagated
        # through from the callback, as it crosses the Python<->C boundary
        # twice. (See GNOME #616279)
        with capture_exceptions() as exc:
            Everything.test_simple_callback(callback)
        self.assertTrue(exc)
        self.assertEqual(exc[0].type, ZeroDivisionError)

    def test_return_value_callback(self):
        TestCallbacks.called = False

        def callback():
            TestCallbacks.called = True
            return 44

        self.assertEqual(Everything.test_callback(callback), 44)
        self.assertTrue(TestCallbacks.called)

    def test_callback_scope_async(self):
        TestCallbacks.called = False
        ud = "Test Value 44"

        def callback(user_data):
            self.assertEqual(user_data, ud)
            TestCallbacks.called = True
            return 44

        if hasattr(sys, "getrefcount"):
            ud_refcount = sys.getrefcount(ud)
            callback_refcount = sys.getrefcount(callback)

        self.assertEqual(Everything.test_callback_async(callback, ud), None)
        # Callback should not have run and the ref count is increased by 1
        self.assertEqual(TestCallbacks.called, False)

        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount + 1)
            self.assertEqual(sys.getrefcount(ud), ud_refcount + 1)

        # test_callback_thaw_async will run the callback previously supplied.
        # references should be auto decremented after this call.
        self.assertEqual(Everything.test_callback_thaw_async(), 44)
        self.assertTrue(TestCallbacks.called)

        # Make sure refcounts are returned to normal
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount)
            self.assertEqual(sys.getrefcount(ud), ud_refcount)

    def test_callback_scope_call_multi(self):
        # This tests a callback that gets called multiple times from a
        # single scope call in python.
        TestCallbacks.called = 0

        def callback():
            TestCallbacks.called += 1
            return TestCallbacks.called

        if hasattr(sys, "getrefcount"):
            refcount = sys.getrefcount(callback)
        result = Everything.test_multi_callback(callback)
        # first callback should give 1, second 2, and the function sums them up
        self.assertEqual(result, 3)
        self.assertEqual(TestCallbacks.called, 2)

        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), refcount)

    def test_callback_scope_call_array(self):
        # This tests a callback that gets called multiple times from a
        # single scope call in python with array arguments
        TestCallbacks.callargs = []

        # FIXME: would be cleaner without the explicit length args:
        # def callback(one, two):
        def callback(one, one_length, two, two_length):
            TestCallbacks.callargs.append((one, two))
            return len(TestCallbacks.callargs)

        if hasattr(sys, "getrefcount"):
            refcount = sys.getrefcount(callback)
        result = Everything.test_array_callback(callback)
        # first callback should give 1, second 2, and the function sums them up
        self.assertEqual(result, 3)
        self.assertEqual(
            TestCallbacks.callargs, [([-1, 0, 1, 2], ["one", "two", "three"])] * 2
        )

        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), refcount)

    def test_callback_scope_call_array_inout(self):
        # This tests a callback that gets called multiple times from a
        # single scope call in python with inout array arguments
        TestCallbacks.callargs = []

        def callback(ints, ints_length):
            TestCallbacks.callargs.append(ints)
            return ints[1:], len(ints[1:])

        if hasattr(sys, "getrefcount"):
            refcount = sys.getrefcount(callback)
        result = Everything.test_array_inout_callback(callback)
        self.assertEqual(TestCallbacks.callargs, [[-2, -1, 0, 1, 2], [-1, 0, 1, 2]])
        # first callback should give 4, second 3
        self.assertEqual(result, 3)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), refcount)

    def test_callback_userdata(self):
        TestCallbacks.called = 0

        def callback(userdata):
            self.assertEqual(userdata, f"Test{TestCallbacks.called:d}")
            TestCallbacks.called += 1
            return TestCallbacks.called

        for i in range(100):
            val = Everything.test_callback_user_data(callback, f"Test{i:d}")
            self.assertEqual(val, i + 1)

        self.assertEqual(TestCallbacks.called, 100)

    def test_callback_userdata_no_user_data(self):
        TestCallbacks.called = 0

        def callback():
            TestCallbacks.called += 1
            return TestCallbacks.called

        for i in range(100):
            val = Everything.test_callback_user_data(callback)
            self.assertEqual(val, i + 1)

        self.assertEqual(TestCallbacks.called, 100)

    def test_callback_userdata_varargs(self):
        TestCallbacks.called = 0
        collected_user_data = []

        def callback(a, b):
            collected_user_data.extend([a, b])
            TestCallbacks.called += 1
            return TestCallbacks.called

        for i in range(10):
            val = Everything.test_callback_user_data(callback, 1, 2)
            self.assertEqual(val, i + 1)

        self.assertEqual(TestCallbacks.called, 10)
        self.assertSequenceEqual(collected_user_data, [1, 2] * 10)

    def test_callback_userdata_as_kwarg_tuple(self):
        TestCallbacks.called = 0
        collected_user_data = []

        def callback(user_data):
            collected_user_data.extend(user_data)
            TestCallbacks.called += 1
            return TestCallbacks.called

        for i in range(10):
            val = Everything.test_callback_user_data(callback, user_data=(1, 2))
            self.assertEqual(val, i + 1)

        self.assertEqual(TestCallbacks.called, 10)
        self.assertSequenceEqual(collected_user_data, [1, 2] * 10)

    def test_callback_user_data_middle_none(self):
        cb_info = {}

        def callback(userdata):
            cb_info["called"] = True
            cb_info["userdata"] = userdata
            return 1

        (y, z, q) = Everything.test_torture_signature_2(
            42, callback, None, "some string", 3
        )
        self.assertEqual(y, 42)
        self.assertEqual(z, 84)
        self.assertEqual(q, 14)
        self.assertTrue(cb_info["called"])
        self.assertEqual(cb_info["userdata"], None)

    def test_callback_user_data_middle_single(self):
        cb_info = {}

        def callback(userdata):
            cb_info["called"] = True
            cb_info["userdata"] = userdata
            return 1

        (y, z, q) = Everything.test_torture_signature_2(
            42, callback, "User Data", "some string", 3
        )
        self.assertEqual(y, 42)
        self.assertEqual(z, 84)
        self.assertEqual(q, 14)
        self.assertTrue(cb_info["called"])
        self.assertEqual(cb_info["userdata"], "User Data")

    def test_callback_user_data_middle_tuple(self):
        cb_info = {}

        def callback(userdata):
            cb_info["called"] = True
            cb_info["userdata"] = userdata
            return 1

        (y, z, q) = Everything.test_torture_signature_2(
            42, callback, (-5, "User Data"), "some string", 3
        )
        self.assertEqual(y, 42)
        self.assertEqual(z, 84)
        self.assertEqual(q, 14)
        self.assertTrue(cb_info["called"])
        self.assertEqual(cb_info["userdata"], (-5, "User Data"))

    def test_async_ready_callback(self):
        TestCallbacks.called = False
        TestCallbacks.main_loop = GLib.MainLoop()

        def callback(obj, result, user_data):
            TestCallbacks.main_loop.quit()
            TestCallbacks.called = True

        Everything.test_async_ready_callback(callback)

        TestCallbacks.main_loop.run()

        self.assertTrue(TestCallbacks.called)

    def test_callback_scope_notified_with_destroy(self):
        TestCallbacks.called = 0
        ud = "Test scope notified data 33"

        def callback(user_data):
            self.assertEqual(user_data, ud)
            TestCallbacks.called += 1
            return 33

        if hasattr(sys, "getrefcount"):
            value_refcount = sys.getrefcount(ud)
            callback_refcount = sys.getrefcount(callback)

        # Callback is immediately called.
        for i in range(100):
            res = Everything.test_callback_destroy_notify(callback, ud)
            self.assertEqual(res, 33)

        self.assertEqual(TestCallbacks.called, 100)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount + 100)
            self.assertEqual(sys.getrefcount(ud), value_refcount + 100)

        # thaw will call the callback again, this time resources should be freed
        self.assertEqual(Everything.test_callback_thaw_notifications(), 33 * 100)
        self.assertEqual(TestCallbacks.called, 200)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount)
            self.assertEqual(sys.getrefcount(ud), value_refcount)

    def test_callback_scope_notified_with_destroy_no_user_data(self):
        TestCallbacks.called = 0

        def callback(user_data):
            self.assertEqual(user_data, None)
            TestCallbacks.called += 1
            return 34

        if hasattr(sys, "getrefcount"):
            callback_refcount = sys.getrefcount(callback)

        # Run with warning as exception
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("error")
            self.assertRaises(
                RuntimeWarning,
                Everything.test_callback_destroy_notify_no_user_data,
                callback,
            )

        self.assertEqual(TestCallbacks.called, 0)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount)

        # Run with warning as warning
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("default")
            # Trigger a warning.
            res = Everything.test_callback_destroy_notify_no_user_data(callback)
            # Verify some things
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, RuntimeWarning))
            self.assertTrue("Callables passed to" in str(w[-1].message))

        self.assertEqual(res, 34)
        self.assertEqual(TestCallbacks.called, 1)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount + 1)

        # thaw will call the callback again,
        # refcount will not go down without user_data parameter
        self.assertEqual(Everything.test_callback_thaw_notifications(), 34)
        self.assertEqual(TestCallbacks.called, 2)
        if hasattr(sys, "getrefcount"):
            self.assertEqual(sys.getrefcount(callback), callback_refcount + 1)

    def test_callback_in_methods(self):
        object_ = Everything.TestObj()

        def callback():
            TestCallbacks.called = True
            return 42

        TestCallbacks.called = False
        object_.instance_method_callback(callback)
        self.assertTrue(TestCallbacks.called)

        TestCallbacks.called = False
        Everything.TestObj.static_method_callback(callback)
        self.assertTrue(TestCallbacks.called)

        def callbackWithUserData(user_data):
            TestCallbacks.called += 1
            return 42

        TestCallbacks.called = 0
        Everything.TestObj.new_callback(callbackWithUserData, None)
        self.assertEqual(TestCallbacks.called, 1)
        # Note: using "new_callback" adds the notification to the same global
        # list as Everything.test_callback_destroy_notify, so thaw the list
        # so we don't get confusion between tests.
        self.assertEqual(Everything.test_callback_thaw_notifications(), 42)
        self.assertEqual(TestCallbacks.called, 2)

    def test_callback_none(self):
        # make sure this doesn't assert or crash
        Everything.test_simple_callback(None)

    def test_callback_gerror(self):
        def callback(error):
            self.assertEqual(error.message, "regression test error")
            self.assertTrue("g-io" in error.domain)
            self.assertEqual(error.code, Gio.IOErrorEnum.NOT_SUPPORTED)
            TestCallbacks.called = True

        TestCallbacks.called = False
        Everything.test_gerror_callback(callback)
        self.assertTrue(TestCallbacks.called)

    def test_callback_null_gerror(self):
        def callback(error):
            self.assertEqual(error, None)
            TestCallbacks.called = True

        TestCallbacks.called = False
        Everything.test_null_gerror_callback(callback)
        self.assertTrue(TestCallbacks.called)

    def test_callback_owned_gerror(self):
        def callback(error):
            self.assertEqual(error.message, "regression test owned error")
            self.assertTrue("g-io" in error.domain)
            self.assertEqual(error.code, Gio.IOErrorEnum.PERMISSION_DENIED)
            TestCallbacks.called = True

        TestCallbacks.called = False
        Everything.test_owned_gerror_callback(callback)
        self.assertTrue(TestCallbacks.called)

    def test_callback_hashtable(self):
        def callback(data):
            self.assertEqual(data, mydict)
            mydict["new"] = 42
            TestCallbacks.called = True

        mydict = {"foo": 1, "bar": 2}
        TestCallbacks.called = False
        Everything.test_hash_table_callback(mydict, callback)
        self.assertTrue(TestCallbacks.called)
        self.assertEqual(mydict, {"foo": 1, "bar": 2, "new": 42})


class TestClosures(unittest.TestCase):
    def test_no_arg(self):
        def callback():
            self.called = True
            return 42

        self.called = False
        result = Everything.test_closure(callback)
        self.assertTrue(self.called)
        self.assertEqual(result, 42)

    def test_int_arg(self):
        def callback(num):
            self.called = True
            return num + 1

        self.called = False
        result = Everything.test_closure_one_arg(callback, 42)
        self.assertTrue(self.called)
        self.assertEqual(result, 43)

    def test_variant(self):
        def callback(variant):
            self.called = True
            if variant is None:
                return None
            self.assertEqual(variant.get_type_string(), "i")
            return GLib.Variant("i", variant.get_int32() + 1)

        self.called = False
        result = Everything.test_closure_variant(callback, GLib.Variant("i", 42))
        self.assertTrue(self.called)
        self.assertEqual(result.get_type_string(), "i")
        self.assertEqual(result.get_int32(), 43)

        self.called = False
        result = Everything.test_closure_variant(callback, None)
        self.assertTrue(self.called)
        self.assertEqual(result, None)

        self.called = False
        self.assertRaises(TypeError, Everything.test_closure_variant, callback, "foo")
        self.assertFalse(self.called)

    def test_variant_wrong_return_type(self):
        def callback(variant):
            return "no_variant"

        with capture_exceptions() as exc:
            # this does not directly raise an exception (see
            # https://bugzilla.gnome.org/show_bug.cgi?id=616279)
            result = Everything.test_closure_variant(callback, GLib.Variant("i", 42))
        # ... but the result shouldn't be a string
        self.assertEqual(result, None)
        # and the error should be shown
        self.assertEqual(len(exc), 1)
        self.assertEqual(exc[0].type, TypeError)
        self.assertTrue("return value" in str(exc[0].value), exc[0].value)


class TestBoxed(unittest.TestCase):
    def test_boxed(self):
        object_ = Everything.TestObj()
        self.assertEqual(object_.props.boxed, None)

        boxed = Everything.TestBoxed()
        boxed.some_int8 = 42
        object_.props.boxed = boxed

        self.assertTrue(isinstance(object_.props.boxed, Everything.TestBoxed))
        self.assertEqual(object_.props.boxed.some_int8, 42)

    def test_boxed_alternative_constructor(self):
        boxed = Everything.TestBoxed.new_alternative_constructor1(5)
        self.assertEqual(boxed.some_int8, 5)

        boxed = Everything.TestBoxed.new_alternative_constructor2(5, 3)
        self.assertEqual(boxed.some_int8, 8)

        boxed = Everything.TestBoxed.new_alternative_constructor3("-3")
        self.assertEqual(boxed.some_int8, -3)

    def test_boxed_equality(self):
        boxed42 = Everything.TestBoxed.new_alternative_constructor1(42)
        boxed5 = Everything.TestBoxed.new_alternative_constructor1(5)
        boxed42_2 = Everything.TestBoxed.new_alternative_constructor2(41, 1)

        self.assertFalse(boxed42.equals(boxed5))
        self.assertTrue(boxed42.equals(boxed42_2))
        self.assertTrue(boxed42_2.equals(boxed42))
        self.assertTrue(boxed42.equals(boxed42))

    def test_boxed_b_constructor(self):
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            boxed = Everything.TestBoxedB(42, 47)
            self.assertTrue(issubclass(warn[0].category, DeprecationWarning))

        self.assertEqual(boxed.some_int8, 0)
        self.assertEqual(boxed.some_long, 0)

    def test_boxed_c_equality(self):
        boxed = Everything.TestBoxedC()
        # TestBoxedC uses refcounting, so we know that
        # the pointer is the same when copied
        copy = boxed.copy()
        self.assertEqual(boxed, copy)
        self.assertNotEqual(id(boxed), id(copy))

    def test_boxed_c_wrapper(self):
        wrapper = Everything.TestBoxedCWrapper()
        obj = wrapper.get()

        # TestBoxedC uses refcounting, so we know that
        # it should be 2 at this point:
        # - one owned by @wrapper
        # - another owned by @obj
        self.assertEqual(obj.refcount, 2)
        del wrapper
        gc.collect()
        self.assertEqual(obj.refcount, 1)

    def test_boxed_c_wrapper_copy(self):
        wrapper = Everything.TestBoxedCWrapper()
        wrapper_copy = wrapper.copy()
        obj = wrapper.get()

        # TestBoxedC uses refcounting, so we know that
        # it should be 3 at this point:
        # - one owned by @wrapper
        # - one owned by @wrapper_copy
        # - another owned by @obj
        self.assertEqual(obj.refcount, 3)
        del wrapper
        gc.collect()
        self.assertEqual(obj.refcount, 2)
        del wrapper_copy
        gc.collect()
        self.assertEqual(obj.refcount, 1)
        del obj
        gc.collect()

    def test_array_fixed_boxed_none_out(self):
        arr = Everything.test_array_fixed_boxed_none_out()
        assert len(arr) == 2
        assert arr[0].refcount == 2
        assert arr[1].refcount == 2

    def test_gvalue_out_boxed(self):
        # As corruption is random data, check several times.
        for i in range(10):
            int8 = random.randint(GLib.MININT8, GLib.MAXINT8)
            assert Everything.test_gvalue_out_boxed(int8).some_int8 == int8

    def test_glist_boxed_none_return(self):
        assert len(Everything.test_glist_boxed_none_return(0)) == 0

        list_ = Everything.test_glist_boxed_none_return(2)
        assert len(list_) == 2
        assert list_[0].refcount == 2
        assert list_[1].refcount == 2

    def test_glist_boxed_full_return(self):
        assert len(Everything.test_glist_boxed_full_return(0)) == 0

        list_ = Everything.test_glist_boxed_full_return(2)
        assert len(list_) == 2
        assert list_[0].refcount == 1
        assert list_[1].refcount == 1


class TestTortureProfile(unittest.TestCase):
    def test_torture_profile(self):
        total_time = 0
        object_ = Everything.TestObj()

        start_time = timeit.default_timer()
        for i in range(10000):
            (_y, _z, _q) = object_.torture_signature_0(5000, "Torture Test 1", 12345)

        end_time = timeit.default_timer()
        delta_time = end_time - start_time
        total_time += delta_time

        start_time = timeit.default_timer()
        for i in range(10000):
            (_y, _z, _q) = Everything.TestObj().torture_signature_0(
                5000, "Torture Test 2", 12345
            )

        end_time = timeit.default_timer()
        delta_time = end_time - start_time
        total_time += delta_time

        start_time = timeit.default_timer()
        for i in range(10000):
            with contextlib.suppress(Exception):
                (_y, _z, _q) = object_.torture_signature_1(
                    5000, "Torture Test 3", 12345
                )
        end_time = timeit.default_timer()
        delta_time = end_time - start_time
        total_time += delta_time

        def callback(userdata):
            return 0

        userdata = [1, 2, 3, 4]
        start_time = timeit.default_timer()
        for i in range(10000):
            (_y, _z, _q) = Everything.test_torture_signature_2(
                5000, callback, userdata, "Torture Test 4", 12345
            )
        end_time = timeit.default_timer()
        delta_time = end_time - start_time
        total_time += delta_time


class TestAdvancedInterfaces(unittest.TestCase):
    def test_array_objs(self):
        obj1, obj2 = Everything.test_array_fixed_out_objects()
        self.assertTrue(isinstance(obj1, Everything.TestObj))
        self.assertTrue(isinstance(obj2, Everything.TestObj))
        self.assertNotEqual(obj1, obj2)

    def test_obj_skip_return_val(self):
        obj = Everything.TestObj()
        ret = obj.skip_return_val(50, 42.0, 60, 2, 3)
        self.assertEqual(len(ret), 3)
        self.assertEqual(ret[0], 51)
        self.assertEqual(ret[1], 61)
        self.assertEqual(ret[2], 32)

    def test_obj_skip_return_val_no_out(self):
        obj = Everything.TestObj()
        # raises an error for 0, succeeds for any other value
        self.assertRaises(GLib.GError, obj.skip_return_val_no_out, 0)

        ret = obj.skip_return_val_no_out(1)
        self.assertEqual(ret, None)



import unittest
import pytest

import gi

try:
    gi.require_foreign("cairo")
    import cairo

    has_cairo = True
except ImportError:
    has_cairo = False

has_region = has_cairo and hasattr(cairo, "Region")

try:
    from gi.repository import Gtk, Gdk

    Gtk, Gdk
except ImportError:
    Gtk = None
    Gdk = None

from gi.repository import GObject, Regress


@unittest.skipUnless(has_cairo, "built without cairo support")
class Test(unittest.TestCase):
    def test_gvalue_converters(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        matrix = cairo.Matrix()
        objects = {
            "CairoContext": context,
            "CairoSurface": surface,
            "CairoFontFace": context.get_font_face(),
            "CairoScaledFont": context.get_scaled_font(),
            "CairoPattern": context.get_source(),
            "CairoMatrix": matrix,
        }
        for type_name, cairo_obj in objects.items():
            gtype = GObject.type_from_name(type_name)
            v = GObject.Value()
            assert v.init(gtype) is None
            assert v.get_value() is None
            v.set_value(None)
            assert v.get_value() is None
            v.set_value(cairo_obj)
            assert v.get_value() == cairo_obj

    def test_cairo_context(self):
        context = Regress.test_cairo_context_full_return()
        self.assertTrue(isinstance(context, cairo.Context))

        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        Regress.test_cairo_context_none_in(context)

    def test_cairo_context_full_in(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        Regress.test_cairo_context_full_in(context)

        with pytest.raises(TypeError):
            Regress.test_cairo_context_full_in(object())

    def test_cairo_context_none_return(self):
        context = Regress.test_cairo_context_none_return()
        self.assertTrue(isinstance(context, cairo.Context))

    def test_cairo_path_full_return(self):
        path = Regress.test_cairo_path_full_return()
        if hasattr(cairo, "Path"):  # pycairo 1.15.1+
            assert isinstance(path, cairo.Path)

    def test_cairo_path_none_in(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        path = context.copy_path()
        Regress.test_cairo_path_none_in(path)
        surface.finish()

        with pytest.raises(TypeError):
            Regress.test_cairo_path_none_in(object())

    def test_cairo_path_full_in_full_return(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        context.move_to(10, 10)
        context.curve_to(10, 10, 3, 4, 5, 6)
        path = context.copy_path()
        new_path = Regress.test_cairo_path_full_in_full_return(path)
        assert list(path) == list(new_path)
        surface.finish()

    def test_cairo_font_options_full_return(self):
        options = Regress.test_cairo_font_options_full_return()
        assert isinstance(options, cairo.FontOptions)

    def test_cairo_font_options_none_return(self):
        options = Regress.test_cairo_font_options_none_return()
        assert isinstance(options, cairo.FontOptions)

    def test_cairo_font_options_full_in(self):
        options = cairo.FontOptions()
        Regress.test_cairo_font_options_full_in(options)

        with pytest.raises(TypeError):
            Regress.test_cairo_font_options_full_in(object())

    def test_cairo_font_options_none_in(self):
        options = cairo.FontOptions()
        Regress.test_cairo_font_options_none_in(options)

    def test_cairo_font_face_full_return(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        font_face = Regress.test_cairo_font_face_full_return(context)

        assert font_face

    def test_cairo_scaled_font_full_return(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        scaled_font = Regress.test_cairo_scaled_font_full_return(context)

        assert scaled_font

    def test_cairo_pattern_full_in(self):
        pattern = cairo.SolidPattern(1, 1, 1, 1)
        Regress.test_cairo_pattern_full_in(pattern)

        with pytest.raises(TypeError):
            Regress.test_cairo_pattern_full_in(object())

    def test_cairo_pattern_none_in(self):
        pattern = cairo.SolidPattern(1, 1, 1, 1)
        Regress.test_cairo_pattern_none_in(pattern)

    def test_cairo_pattern_full_return(self):
        pattern = Regress.test_cairo_pattern_full_return()
        self.assertTrue(isinstance(pattern, cairo.Pattern))
        self.assertTrue(isinstance(pattern, cairo.SolidPattern))

    def test_cairo_pattern_none_return(self):
        pattern = Regress.test_cairo_pattern_none_return()
        self.assertTrue(isinstance(pattern, cairo.Pattern))
        self.assertTrue(isinstance(pattern, cairo.SolidPattern))

    def test_cairo_region_full_in(self):
        region = cairo.Region()
        Regress.test_cairo_region_full_in(region)

        with pytest.raises(TypeError):
            Regress.test_cairo_region_full_in(object())

    def test_cairo_matrix_none_in(self):
        matrix = cairo.Matrix()
        Regress.test_cairo_matrix_none_in(matrix)

        with pytest.raises(TypeError):
            Regress.test_cairo_matrix_none_in(object())

    def test_cairo_matrix_none_return(self):
        matrix = Regress.test_cairo_matrix_none_return()
        assert matrix == cairo.Matrix()

    def test_cairo_matrix_out_caller_allocates(self):
        matrix = Regress.test_cairo_matrix_out_caller_allocates()
        assert matrix == cairo.Matrix()

    def test_cairo_surface(self):
        surface = Regress.test_cairo_surface_none_return()
        self.assertTrue(isinstance(surface, cairo.ImageSurface))
        self.assertTrue(isinstance(surface, cairo.Surface))
        self.assertEqual(surface.get_format(), cairo.FORMAT_ARGB32)
        self.assertEqual(surface.get_width(), 10)
        self.assertEqual(surface.get_height(), 10)

        surface = Regress.test_cairo_surface_full_return()
        self.assertTrue(isinstance(surface, cairo.ImageSurface))
        self.assertTrue(isinstance(surface, cairo.Surface))
        self.assertEqual(surface.get_format(), cairo.FORMAT_ARGB32)
        self.assertEqual(surface.get_width(), 10)
        self.assertEqual(surface.get_height(), 10)

        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        Regress.test_cairo_surface_none_in(surface)

        surface = Regress.test_cairo_surface_full_out()
        self.assertTrue(isinstance(surface, cairo.ImageSurface))
        self.assertTrue(isinstance(surface, cairo.Surface))
        self.assertEqual(surface.get_format(), cairo.FORMAT_ARGB32)
        self.assertEqual(surface.get_width(), 10)
        self.assertEqual(surface.get_height(), 10)

    def test_cairo_surface_full_in(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        Regress.test_cairo_surface_full_in(surface)

        with pytest.raises(TypeError):
            Regress.test_cairo_surface_full_in(object())

    def test_require_foreign(self):
        self.assertEqual(gi.require_foreign("cairo"), None)
        self.assertEqual(gi.require_foreign("cairo", "Context"), None)
        self.assertRaises(ImportError, gi.require_foreign, "invalid_module")
        self.assertRaises(
            ImportError, gi.require_foreign, "invalid_module", "invalid_symbol"
        )
        self.assertRaises(ImportError, gi.require_foreign, "cairo", "invalid_symbol")


@unittest.skipUnless(has_cairo, "built without cairo support")
@unittest.skipUnless(has_region, "built without cairo.Region support")
@unittest.skipUnless(Gdk, "Gdk not available")
class TestRegion(unittest.TestCase):
    def test_region_to_py(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        context.paint()
        region = Gdk.cairo_region_create_from_surface(surface)
        r = region.get_extents()
        self.assertEqual((r.height, r.width), (10, 10))

    def test_region_from_py(self):
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        region = cairo.Region(cairo.RectangleInt(0, 0, 42, 42))
        Gdk.cairo_region(context, region)
        self.assertTrue("42" in repr(list(context.copy_path())))


@unittest.skipUnless(has_cairo, "built without cairo support")
@unittest.skipUnless(Gtk, "Gtk not available")
class TestPango(unittest.TestCase):
    def test_cairo_font_options(self):
        window = Gtk.Window()
        if Gtk._version == "4.0":
            window.set_font_options(cairo.FontOptions())
            font_opts = window.get_font_options()
        else:
            screen = window.get_screen()
            font_opts = screen.get_font_options()
        assert font_opts is not None
        self.assertTrue(isinstance(font_opts.get_subpixel_order(), int))


if has_cairo:
    from gi.repository import cairo as CairoGObject

    # Use PyGI signals to test non-introspected foreign marshaling.
    class CairoSignalTester(GObject.Object):
        sig_context = GObject.Signal(arg_types=[CairoGObject.Context])
        sig_surface = GObject.Signal(arg_types=[CairoGObject.Surface])
        sig_font_face = GObject.Signal(arg_types=[CairoGObject.FontFace])
        sig_scaled_font = GObject.Signal(arg_types=[CairoGObject.ScaledFont])
        sig_pattern = GObject.Signal(arg_types=[CairoGObject.Pattern])


@unittest.skipUnless(has_cairo, "built without cairo support")
class TestSignalMarshaling(unittest.TestCase):
    # Tests round tripping of cairo objects through non-introspected signals.

    def setUp(self):
        self.surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        self.context = cairo.Context(self.surface)
        self.tester = CairoSignalTester()

    def pass_object_through_signal(self, obj, signal):
        """Pass the given `obj` through the `signal` emission storing the
        `obj` passed through the signal and returning it.
        """
        passthrough_result = []

        def callback(instance, passthrough):
            passthrough_result.append(passthrough)

        signal.connect(callback)
        signal.emit(obj)

        return passthrough_result[0]

    def test_context(self):
        result = self.pass_object_through_signal(self.context, self.tester.sig_context)
        self.assertTrue(isinstance(result, cairo.Context))

        with pytest.raises(TypeError):
            self.pass_object_through_signal(object(), self.tester.sig_context)

    def test_surface(self):
        result = self.pass_object_through_signal(self.surface, self.tester.sig_surface)
        self.assertTrue(isinstance(result, cairo.Surface))

    def test_font_face(self):
        font_face = self.context.get_font_face()
        result = self.pass_object_through_signal(font_face, self.tester.sig_font_face)
        self.assertTrue(isinstance(result, cairo.FontFace))

        with pytest.raises(TypeError):
            self.pass_object_through_signal(object(), self.tester.sig_font_face)

    def test_scaled_font(self):
        scaled_font = cairo.ScaledFont(
            self.context.get_font_face(),
            cairo.Matrix(),
            cairo.Matrix(),
            self.context.get_font_options(),
        )
        result = self.pass_object_through_signal(
            scaled_font, self.tester.sig_scaled_font
        )
        self.assertTrue(isinstance(result, cairo.ScaledFont))

        with pytest.raises(TypeError):
            result = self.pass_object_through_signal(
                object(), self.tester.sig_scaled_font
            )

    def test_pattern(self):
        pattern = cairo.SolidPattern(1, 1, 1, 1)
        result = self.pass_object_through_signal(pattern, self.tester.sig_pattern)
        self.assertTrue(isinstance(result, cairo.Pattern))
        self.assertTrue(isinstance(result, cairo.SolidPattern))

        with pytest.raises(TypeError):
            result = self.pass_object_through_signal(object(), self.tester.sig_pattern)



import unittest

from gi.repository import GObject

import testhelper


TestInterface = GObject.GType.from_name("TestInterface")


class TestUnknown(unittest.TestCase):
    def test_unknown_interface(self):
        obj = testhelper.get_unknown()
        TestUnknownGType = GObject.GType.from_name("TestUnknown")
        TestUnknown = GObject.new(TestUnknownGType).__class__
        assert isinstance(obj, testhelper.Interface)
        assert isinstance(obj, TestUnknown)

    def test_property(self):
        obj = testhelper.get_unknown()
        self.assertEqual(obj.get_property("some-property"), None)
        obj.set_property("some-property", "foo")

    def test_unknown_property(self):
        obj = testhelper.get_unknown()
        self.assertRaises(TypeError, obj.get_property, "unknown")
        self.assertRaises(TypeError, obj.set_property, "unknown", "1")



import contextlib
import os
import sys
import unittest
import os.path
import warnings
import subprocess

import pytest
from gi.repository import GLib
from gi import PyGIDeprecationWarning


PLATFORM_SPLIT_VERSION = (2, 88)


GLibUnix = None
with contextlib.suppress(ImportError):
    from gi.repository import GLibUnix


class TestGLib(unittest.TestCase):
    @pytest.mark.xfail()
    def test_pytest_capture_error_in_closure(self):
        # this test is supposed to fail
        ml = GLib.MainLoop()

        def callback():
            ml.quit()
            raise Exception("expected")

        GLib.idle_add(callback)
        ml.run()

    @unittest.skipIf(os.name == "nt", "no bash on Windows")
    def test_find_program_in_path(self):
        bash_path = GLib.find_program_in_path("bash")
        self.assertTrue(bash_path.endswith(os.path.sep + "bash"))
        self.assertTrue(os.path.exists(bash_path))

        self.assertEqual(GLib.find_program_in_path("non existing"), None)

    def test_markup_escape_text(self):
        self.assertEqual(GLib.markup_escape_text("a&bä"), "a&amp;bä")
        self.assertEqual(GLib.markup_escape_text(b"a&b\x05"), "a&amp;b&#x5;")

        # with explicit length argument
        self.assertEqual(GLib.markup_escape_text(b"a\x05\x01\x02", 2), "a&#x5;")

    def test_progname(self):
        GLib.set_prgname("moo")
        self.assertEqual(GLib.get_prgname(), "moo")

    def test_appname(self):
        GLib.set_application_name("moo")
        self.assertEqual(GLib.get_application_name(), "moo")

    def test_xdg_dirs(self):
        d = GLib.get_user_data_dir()
        self.assertTrue(os.path.sep in d, d)
        d = GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP)
        self.assertTrue(os.path.sep in d, d)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            # also works with backwards compatible enum names
            self.assertEqual(
                GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_MUSIC),
                GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC),
            )

        for d in GLib.get_system_config_dirs():
            self.assertTrue(os.path.sep in d, d)
        for d in GLib.get_system_data_dirs():
            self.assertTrue(isinstance(d, str), d)

    def test_main_depth(self):
        self.assertEqual(GLib.main_depth(), 0)

    def test_filenames(self):
        self.assertEqual(GLib.filename_display_name("foo"), "foo")
        self.assertEqual(GLib.filename_display_basename("bar/foo"), "foo")

        def glibfsencode(f):
            # the annotations of filename_from_utf8() was changed in
            # https://bugzilla.gnome.org/show_bug.cgi?id=756128
            if isinstance(f, bytes):
                return f
            if os.name == "nt":
                return f.encode("utf-8", "surrogatepass")
            return os.fsencode(f)

        # this is locale dependent, so we cannot completely verify the result
        res = GLib.filename_from_utf8("aäb")
        res = glibfsencode(res)
        self.assertTrue(isinstance(res, bytes))
        self.assertGreaterEqual(len(res), 3)

        # with explicit length argument
        res = GLib.filename_from_utf8("aäb", 1)
        res = glibfsencode(res)
        self.assertEqual(res, b"a")

    def test_uri_extract(self):
        res = GLib.uri_list_extract_uris("""# some comment
http://example.com
https://my.org/q?x=1&y=2
            http://gnome.org/new""")
        self.assertEqual(
            res,
            ["http://example.com", "https://my.org/q?x=1&y=2", "http://gnome.org/new"],
        )

    def test_current_time(self):
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            tm = GLib.get_current_time()
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        self.assertTrue(isinstance(tm, float))
        self.assertGreater(tm, 1350000000.0)

    @unittest.skipIf(sys.platform == "darwin", "fails on OSX")
    def test_main_loop(self):
        # note we do not test run() here, as we use this in countless other
        # tests
        ml = GLib.MainLoop()
        self.assertFalse(ml.is_running())

        context = ml.get_context()
        self.assertEqual(context, GLib.MainContext.default())
        self.assertTrue(context.is_owner() in [True, False])
        self.assertTrue(context.pending() in [True, False])
        self.assertFalse(context.iteration(False))

    def test_main_loop_with_context(self):
        context = GLib.MainContext()
        ml = GLib.MainLoop(context)
        self.assertFalse(ml.is_running())
        self.assertEqual(ml.get_context(), context)

    def test_main_context(self):
        # constructor
        context = GLib.MainContext()
        self.assertTrue(context.is_owner() in [True, False])
        self.assertFalse(context.pending())
        self.assertFalse(context.iteration(False))

        # GLib API
        context = GLib.MainContext.default()
        self.assertTrue(context.is_owner() in [True, False])
        self.assertTrue(context.pending() in [True, False])
        self.assertTrue(context.iteration(False) in [True, False])

        # backwards compatible API
        context = GLib.main_context_default()
        self.assertTrue(context.is_owner() in [True, False])
        self.assertTrue(context.pending() in [True, False])
        self.assertTrue(context.iteration(False) in [True, False])

    def test_main_context_query(self):
        context = GLib.MainContext()
        assert context.acquire()

        timeout_msec, fds = context.query(0)

        assert timeout_msec == 0
        assert len(fds) == 1

    @unittest.skipIf(os.name == "nt", "hangs")
    def test_io_add_watch_no_data(self):
        (r, w) = os.pipe()
        call_data = []

        def cb(fd, condition):
            call_data.append((fd, condition, os.read(fd, 1)))
            if len(call_data) == 2:
                ml.quit()
            return True

        # io_add_watch() takes an IOChannel, calling with an fd is deprecated
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            GLib.io_add_watch(r, GLib.IOCondition.IN, cb, priority=GLib.PRIORITY_HIGH)
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        ml = GLib.MainLoop()
        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(
            call_data, [(r, GLib.IOCondition.IN, b"a"), (r, GLib.IOCondition.IN, b"b")]
        )

    @unittest.skipIf(os.name == "nt", "hangs")
    def test_io_add_watch_with_data(self):
        (r, w) = os.pipe()
        call_data = []

        def cb(fd, condition, data):
            call_data.append((fd, condition, os.read(fd, 1), data))
            if len(call_data) == 2:
                ml.quit()
            return True

        # io_add_watch() takes an IOChannel, calling with an fd is deprecated
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            GLib.io_add_watch(
                r, GLib.IOCondition.IN, cb, "moo", priority=GLib.PRIORITY_HIGH
            )
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        def write():
            os.write(w, b"a")
            GLib.idle_add(lambda: os.write(w, b"b") and False)

        ml = GLib.MainLoop()
        GLib.idle_add(write)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(
            call_data,
            [
                (r, GLib.IOCondition.IN, b"a", "moo"),
                (r, GLib.IOCondition.IN, b"b", "moo"),
            ],
        )

    @unittest.skipIf(os.name == "nt", "hangs")
    def test_io_add_watch_with_multiple_data(self):
        (r, w) = os.pipe()
        call_data = []

        def cb(fd, condition, *user_data):
            call_data.append((fd, condition, os.read(fd, 1), user_data))
            ml.quit()
            return True

        # io_add_watch() takes an IOChannel, calling with an fd is deprecated
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            GLib.io_add_watch(r, GLib.IOCondition.IN, cb, "moo", "foo")
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        ml = GLib.MainLoop()
        GLib.idle_add(lambda: os.write(w, b"a") and False)
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        self.assertEqual(call_data, [(r, GLib.IOCondition.IN, b"a", ("moo", "foo"))])

    @unittest.skipIf(sys.platform == "darwin", "fails")
    @unittest.skipIf(os.name == "nt", "no shell on Windows")
    def test_io_add_watch_pyfile(self):
        call_data = []

        cmd = subprocess.Popen(
            "echo hello; echo world", shell=True, bufsize=0, stdout=subprocess.PIPE
        )

        def cb(file, condition):
            call_data.append((file, condition, file.readline()))
            if len(call_data) == 2:
                # avoid having to wait for the full timeout
                ml.quit()
            return True

        # io_add_watch() takes an IOChannel, calling with a Python file is deprecated
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            GLib.io_add_watch(cmd.stdout, GLib.IOCondition.IN, cb)
            self.assertTrue(issubclass(warn[0].category, PyGIDeprecationWarning))

        ml = GLib.MainLoop()
        GLib.timeout_add(2000, ml.quit)
        ml.run()

        cmd.wait()

        self.assertEqual(
            call_data,
            [
                (cmd.stdout, GLib.IOCondition.IN, b"hello\n"),
                (cmd.stdout, GLib.IOCondition.IN, b"world\n"),
            ],
        )

    def test_glib_version(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            (major, minor, micro) = GLib.glib_version
            self.assertGreaterEqual(major, 2)
            self.assertGreaterEqual(minor, 0)
            self.assertGreaterEqual(micro, 0)

    def test_pyglib_version(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            (major, minor, micro) = GLib.pyglib_version
            self.assertGreaterEqual(major, 3)
            self.assertGreaterEqual(minor, 0)
            self.assertGreaterEqual(micro, 0)

    def test_timezone_constructor(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            timezone = GLib.TimeZone("+05:21")
            self.assertEqual(timezone.get_offset(0), ((5 * 60) + 21) * 60)

    def test_source_attach_implicit_context(self):
        context = GLib.MainContext.default()
        source = GLib.Idle()
        source_id = source.attach()
        self.assertEqual(context, source.get_context())
        self.assertTrue(GLib.Source.remove(source_id))


class TestGLibPlatform(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # TODO: Make this more dynamic in case we'll support Win32
        cls.platform_glib = GLibUnix

        if cls.platform_glib:
            cls.platform_full_name = cls.platform_glib._namespace
            cls.platform_name = f"{cls.platform_full_name[len(GLib._namespace) :]}"

        unittest.TestCase.setUpClass()

    def assertPlatformTypeFallback(self, symbol_name):
        self.assertIn(symbol_name, dir(GLib))

        platform_symbol_name = symbol_name
        if symbol_name.startswith(f"{self.platform_name}"):
            platform_symbol_name = symbol_name[len(self.platform_name) :]
        elif symbol_name.startswith(f"{self.platform_name.lower()}_"):
            platform_symbol_name = symbol_name[len(self.platform_name) + 1 :]

        with warnings.catch_warnings(record=True) as warn:
            symbol = getattr(GLib, symbol_name)
            self.assertIsNotNone(symbol)

            platform_symbol = getattr(self.platform_glib, platform_symbol_name)
            self.assertIsNotNone(platform_symbol)
            self.maybeAssertDeprecationWarning(warn, symbol_name, platform_symbol_name)

            if (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) >= PLATFORM_SPLIT_VERSION:
                self.assertEqual(platform_symbol, symbol)

    def maybeAssertDeprecationWarning(self, warn, symbol_name, new_symbol_name):
        if (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) >= PLATFORM_SPLIT_VERSION:
            self.assertDeprecationWarning(
                warn,
                "GLib",
                self.platform_full_name,
                symbol_name,
                new_symbol_name,
            )

    def assertDeprecationWarning(
        self, warn, namespace, new_namespace, symbol_name, new_symbol_name
    ):
        self.assertGreaterEqual(len(warn), 1)
        self.assertTrue(
            any(issubclass(w.category, PyGIDeprecationWarning) for w in warn)
        )

        expected_message = (
            f"{namespace}.{symbol_name} is deprecated; "
            + f"use {new_namespace}.{new_symbol_name} instead"
        )
        if len(warn) == 1:
            self.assertEqual(expected_message, str(warn[0].message))
            return

        self.assertIn(
            expected_message,
            [str(w.message) for w in warn],
        )

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_pipe(self):
        self.assertPlatformTypeFallback("UnixPipe")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_pipe_end(self):
        self.assertPlatformTypeFallback("UnixPipeEnd")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_closefrom(self):
        self.assertPlatformTypeFallback("closefrom")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_error_quark(self):
        self.assertPlatformTypeFallback("unix_error_quark")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_fd_add_full(self):
        self.assertPlatformTypeFallback("unix_fd_add_full")

    @unittest.skipIf(
        GLibUnix is None or (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) < (2, 87),
        "Not supported",
    )
    def test_fallback_unix_fd_query_path(self):
        self.assertPlatformTypeFallback("unix_fd_query_path")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_fd_source_new(self):
        self.assertPlatformTypeFallback("unix_fd_source_new")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_fdwalk_set_cloexec(self):
        self.assertPlatformTypeFallback("fdwalk_set_cloexec")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_get_passwd_entry(self):
        self.assertPlatformTypeFallback("unix_get_passwd_entry")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_open_pipe(self):
        self.assertPlatformTypeFallback("unix_open_pipe")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_set_fd_nonblocking(self):
        self.assertPlatformTypeFallback("unix_set_fd_nonblocking")

    @unittest.skipIf(
        GLibUnix is None
        or (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) < PLATFORM_SPLIT_VERSION,
        "Not supported",
    )
    def test_fallback_unix_signal_add(self):
        self.assertPlatformTypeFallback("unix_signal_add")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_signal_add_full(self):
        self.assertPlatformTypeFallback(
            "unix_signal_add"
            if "signal_add" in dir(GLibUnix)
            else "unix_signal_add_full"
        )

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_fallback_unix_signal_source_new(self):
        self.assertPlatformTypeFallback("unix_signal_source_new")

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_glib_unix_pipe(self):
        pipe = GLibUnix.Pipe()
        self.assertEqual(pipe.fds, [0, 0])

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_glib_unix_pipe_fallback(self):
        with warnings.catch_warnings(record=True) as warn:
            pipe = GLib.UnixPipe()
            self.maybeAssertDeprecationWarning(warn, "UnixPipe", "Pipe")
        self.assertEqual(pipe.fds, [0, 0])

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_glib_unix_error_quark(self):
        self.assertGreater(GLibUnix.error_quark(), 0)

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_glib_unix_error_quark_fallback(self):
        with warnings.catch_warnings(record=True) as warn:
            self.assertGreater(GLib.unix_error_quark(), 0)
            self.maybeAssertDeprecationWarning(warn, "unix_error_quark", "error_quark")

    @unittest.skipIf(
        GLibUnix is None
        or (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) < PLATFORM_SPLIT_VERSION,
        "Not supported",
    )
    def test_glib_unix_signal_add_full_deprecation(self):
        with warnings.catch_warnings(record=True) as warn:
            self.assertIsNotNone(GLibUnix.signal_add_full)
            self.assertDeprecationWarning(
                warn, "GLibUnix", "GLibUnix", "signal_add_full", "signal_add"
            )

    @unittest.skipIf(GLibUnix is None, "Not supported")
    def test_glib_unix_signal_add_full_wrapper_deprecation(self):
        with warnings.catch_warnings(record=True) as warn:
            self.assertIsNotNone(GLib.unix_signal_add_full)
            new_namespace = "GLibUnix" if "signal_add" in dir(GLibUnix) else "GLib"
            new_symbol = (
                "signal_add" if "signal_add" in dir(GLibUnix) else "unix_signal_add"
            )
            self.assertDeprecationWarning(
                warn, "GLib", new_namespace, "unix_signal_add_full", new_symbol
            )



import unittest
import weakref
import gc
import sys
import warnings

from gi import PyGIDeprecationWarning
from gi.repository import GObject
from gi.repository import GIMarshallingTests
import testhelper

try:
    from gi.repository import Regress
except ImportError:
    Regress = None


class StrongRef:
    # A class that behaves like weakref.ref but holds a strong reference.
    # This allows re-use of the VFuncsBase by swapping out the ObjectRef
    # class var with either weakref.ref or StrongRef.

    def __init__(self, obj):
        self.obj = obj

    def __call__(self):
        return self.obj


class VFuncsBase(GIMarshallingTests.Object):
    # Class which generically implements the vfuncs used for reference counting tests
    # in a way that can be easily sub-classed and modified.

    #: Object type used by this class for testing
    #: This can be GObject.Object or GObject.InitiallyUnowned
    Object = GObject.Object

    #: Reference type used by this class for holding refs to in/out objects.
    #: This can be set to weakref.ref or StrongRef
    ObjectRef = weakref.ref

    def __init__(self):
        super().__init__()

        #: Hold ref of input or output python wrappers
        self.object_ref = None

        #: store grefcount of input object
        self.in_object_grefcount = None
        self.in_object_is_floating = None

    def do_vfunc_return_object_transfer_none(self):
        # Return an object but keep a python reference to it.
        obj = self.Object()
        self.object_ref = self.ObjectRef(obj)
        return obj

    def do_vfunc_return_object_transfer_full(self):
        # Return an object and hand off the reference to the caller.
        obj = self.Object()
        self.object_ref = self.ObjectRef(obj)
        return obj

    def do_vfunc_out_object_transfer_none(self):
        # Same as do_vfunc_return_object_transfer_none but the pygi
        # internals convert the return here into an out arg.
        obj = self.Object()
        self.object_ref = self.ObjectRef(obj)
        return obj

    def do_vfunc_out_object_transfer_full(self):
        # Same as do_vfunc_return_object_transfer_full but the pygi
        # internals convert the return here into an out arg.
        obj = self.Object()
        self.object_ref = self.ObjectRef(obj)
        return obj

    def do_vfunc_in_object_transfer_none(self, obj):
        # 'obj' will have a python wrapper as well as still held
        # by the caller.
        self.object_ref = self.ObjectRef(obj)
        self.in_object_grefcount = obj.__grefcount__
        self.in_object_is_floating = obj.is_floating()

    def do_vfunc_in_object_transfer_full(self, obj):
        # 'obj' will now be owned by the Python GObject wrapper.
        # When obj goes out of scope and is collected, the GObject
        # should also be fully released.
        self.object_ref = self.ObjectRef(obj)
        self.in_object_grefcount = obj.__grefcount__
        self.in_object_is_floating = obj.is_floating()


class TestVFuncsWithObjectArg(unittest.TestCase):
    # Basic set of tests which work on non-floating objects which python does
    # not keep an additional reference of.

    class VFuncs(VFuncsBase):
        # Object for testing non-floating objects without holding any refs.
        Object = GObject.Object
        ObjectRef = weakref.ref

    def test_vfunc_self_arg_ref_count(self):
        # Check to make sure vfunc "self" arguments don't leak.
        vfuncs = self.VFuncs()
        vfuncs_ref = weakref.ref(vfuncs)
        vfuncs.get_ref_info_for_vfunc_return_object_transfer_full()  # Use any vfunc to test this.

        gc.collect()
        if hasattr(sys, "getrefcount"):
            self.assertEqual(
                sys.getrefcount(vfuncs), 2 if sys.version_info < (3, 14) else 1
            )
        self.assertEqual(vfuncs.__grefcount__, 1)

        del vfuncs
        gc.collect()
        self.assertIsNone(vfuncs_ref())

    def test_vfunc_return_object_transfer_none(self):
        # This tests a problem case where the vfunc returns a GObject owned solely by Python
        # but the argument is marked as transfer-none.
        # In this case pygobject marshaling adds an additional ref and gives a warning
        # of a potential leak. If this occures it is really a bug in the underlying library
        # but pygobject tries to react to this in a reasonable way.
        vfuncs = self.VFuncs()
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            ref_count, is_floating = (
                vfuncs.get_ref_info_for_vfunc_return_object_transfer_none()
            )

        # The ref count of the GObject returned to the caller (get_ref_info_for_vfunc_return_object_transfer_none)
        # should be a single ref
        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_out_object_transfer_none(self):
        # Same as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            ref_count, is_floating = (
                vfuncs.get_ref_info_for_vfunc_out_object_transfer_none()
            )

        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_return_object_transfer_full(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_full()
        )

        # The vfunc caller receives full ownership of a single ref which should not
        # be floating.
        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_out_object_transfer_full(self):
        # Same as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_full()
        )

        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_in_object_transfer_none(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_none(
            self.VFuncs.Object
        )

        gc.collect()
        self.assertEqual(vfuncs.in_object_grefcount, 2)  # initial + python wrapper
        self.assertFalse(vfuncs.in_object_is_floating)

        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)  # ensure python wrapper released
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_in_object_transfer_full(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_full(
            self.VFuncs.Object
        )

        gc.collect()

        # python wrapper should take sole ownership of the gobject
        self.assertEqual(vfuncs.in_object_grefcount, 1)
        self.assertFalse(vfuncs.in_object_is_floating)

        # ensure python wrapper took ownership and released, after vfunc was complete
        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 0)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)


class TestVFuncsWithFloatingArg(unittest.TestCase):
    # All tests here work with a floating object by using InitiallyUnowned as the argument
    # The floating object is sunk as soon as possible.

    class VFuncs(VFuncsBase):
        # Object for testing non-floating objects without holding any refs.
        Object = GObject.InitiallyUnowned
        ObjectRef = weakref.ref

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "refcount specific")
    def test_vfunc_return_object_transfer_none_with_floating(self):
        # See TestVFuncWithObjectArg.test_vfunc_return_object_transfer_none for an explaination.
        vfuncs = self.VFuncs()
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            ref_count, is_floating = (
                vfuncs.get_ref_info_for_vfunc_return_object_transfer_none()
            )

        # The ref count of the GObject returned to the caller (get_ref_info_for_vfunc_return_object_transfer_none)
        # should be a single ref
        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    @unittest.skipUnless(hasattr(sys, "getrefcount"), "refcount specific")
    def test_vfunc_out_object_transfer_none_with_floating(self):
        # Same as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always")
            ref_count, is_floating = (
                vfuncs.get_ref_info_for_vfunc_out_object_transfer_none()
            )

        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_return_object_transfer_full_with_floating(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_full()
        )

        # The vfunc caller receives full ownership of a single ref.
        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_out_object_transfer_full_with_floating(self):
        # Same as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_full()
        )

        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)

    def test_vfunc_in_object_transfer_none_with_floating(self):
        vfuncs = self.VFuncs()

        def ref(obj):
            testhelper.force_g_object_ref(obj)

            # NB. weak reference the underlying GObject!
            return obj.weak_ref()

        vfuncs.ObjectRef = ref

        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_none(
            self.VFuncs.Object
        )
        gc.collect()

        # New: all objects are "sink"'ed (sunken?). Avoid using floating objects altogether.

        # python wrapper should maintain the object as floating and add an additional ref
        # So, two plus the one we added explicitly.
        self.assertEqual(vfuncs.in_object_grefcount, 2 + 1)
        self.assertFalse(vfuncs.in_object_is_floating)

        # Special case for PyPy: the GC has not run, so we have registered an extra reference
        # Add a call to `PyGC_Collect ()` at the end of `_pygi_closure_handle` to check.
        pypy_needs_gc = int(sys.implementation.name == "pypy")
        self.assertEqual(ref_count, 2 + pypy_needs_gc)
        self.assertFalse(is_floating)

        # There are two references now, one explicit and one from the wrapper
        self.assertEqual(vfuncs.object_ref().__grefcount__, 2)
        self.assertFalse(vfuncs.object_ref().is_floating())

        testhelper.force_g_object_unref(vfuncs.object_ref())

        gc.collect()
        assert vfuncs.object_ref() is None

    def test_vfunc_in_object_transfer_full_with_floating(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_full(
            self.VFuncs.Object
        )

        gc.collect()

        # python wrapper sinks and owns the gobject
        self.assertEqual(vfuncs.in_object_grefcount, 1)
        self.assertFalse(vfuncs.in_object_is_floating)

        # ensure python wrapper took ownership and released
        if hasattr(sys, "getrefcount"):
            self.assertEqual(ref_count, 0)
        self.assertFalse(is_floating)

        gc.collect()
        self.assertTrue(vfuncs.object_ref() is None)


class TestVFuncsWithHeldObjectArg(unittest.TestCase):
    # Same tests as TestVFuncsWithObjectArg except we hold
    # onto the python object reference in all cases.

    class VFuncs(VFuncsBase):
        # Object for testing non-floating objects with a held ref.
        Object = GObject.Object
        ObjectRef = StrongRef

    def test_vfunc_return_object_transfer_none_with_held_object(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_none()
        )

        # Python holds the single gobject ref in 'vfuncs.object_ref'
        # Because of this, we do not expect a floating ref or a ref increase.
        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        # The actual grefcount should stay at 1 even after the vfunc return.
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)
        self.assertFalse(vfuncs.in_object_is_floating)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_out_object_transfer_none_with_held_object(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_none()
        )

        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)
        self.assertFalse(vfuncs.in_object_is_floating)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_return_object_transfer_full_with_held_object(self):
        # The vfunc caller receives full ownership which should not
        # be floating. However, the held python wrapper also has a ref.

        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_full()
        )

        # Ref count from the perspective of C after the vfunc is called
        # The vfunc caller receives a new reference which should not
        # be floating. However, the held python wrapper also has a ref.
        self.assertEqual(ref_count, 2)
        self.assertFalse(is_floating)

        # Current ref count
        # The vfunc caller should have decremented its reference.
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_out_object_transfer_full_with_held_object(self):
        # Same test as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_full()
        )

        # Ref count from the perspective of C after the vfunc is called
        # The vfunc caller receives a new reference which should not
        # be floating. However, the held python wrapper also has a ref.
        self.assertEqual(ref_count, 2)
        self.assertFalse(is_floating)

        # Current ref count
        # The vfunc caller should have decremented its reference.
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_in_object_transfer_none_with_held_object(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_none(
            self.VFuncs.Object
        )

        gc.collect()

        # Ref count inside vfunc from the perspective of Python
        self.assertEqual(vfuncs.in_object_grefcount, 2)  # initial + python wrapper
        self.assertFalse(vfuncs.in_object_is_floating)

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(ref_count, 2)  # kept after vfunc + held python wrapper
        self.assertFalse(is_floating)

        # Current ref count after C cleans up its reference
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_in_object_transfer_full_with_held_object(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_full(
            self.VFuncs.Object
        )

        gc.collect()

        # Ref count inside vfunc from the perspective of Python
        self.assertEqual(
            vfuncs.in_object_grefcount, 1
        )  # python wrapper takes ownership of the gobject
        self.assertFalse(vfuncs.in_object_is_floating)

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        # Current ref count
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)


class TestVFuncsWithHeldFloatingArg(unittest.TestCase):
    # Tests for a floating object which we hold a reference to the python wrapper
    # on the VFuncs test class.

    class VFuncs(VFuncsBase):
        # Object for testing floating objects with a held ref.
        Object = GObject.InitiallyUnowned
        ObjectRef = StrongRef

    def test_vfunc_return_object_transfer_none_with_held_floating(self):
        # Python holds onto the wrapper which basically means the floating ref
        # should also be owned by python.

        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_none()
        )

        # This is a borrowed ref from what is held in python.
        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        # The actual grefcount should stay at 1 even after the vfunc return.
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_out_object_transfer_none_with_held_floating(self):
        # Same as above

        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_none()
        )

        self.assertEqual(ref_count, 1)
        self.assertFalse(is_floating)

        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_return_object_transfer_full_with_held_floating(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_return_object_transfer_full()
        )

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(ref_count, 2)
        self.assertFalse(is_floating)

        # Current ref count
        # vfunc wrapper destroyes ref it was given
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref)
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_out_object_transfer_full_with_held_floating(self):
        # Same test as above except uses out arg instead of return
        vfuncs = self.VFuncs()
        ref_count, is_floating = (
            vfuncs.get_ref_info_for_vfunc_out_object_transfer_full()
        )

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(ref_count, 2)
        self.assertFalse(is_floating)

        # Current ref count
        # vfunc wrapper destroyes ref it was given
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_in_floating_transfer_none_with_held_floating(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_none(
            self.VFuncs.Object
        )
        gc.collect()

        # Ref count inside vfunc from the perspective of Python
        self.assertFalse(vfuncs.in_object_is_floating)
        self.assertEqual(
            vfuncs.in_object_grefcount, 2
        )  # python wrapper sinks and owns the gobject

        # Ref count from the perspective of C after the vfunc is called
        self.assertFalse(is_floating)
        self.assertEqual(ref_count, 2)  # floating + held by wrapper

        # Current ref count after C cleans up its reference
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        # Our reference is still floating at this point ...
        self.assertFalse(vfuncs.object_ref().is_floating())
        # ... usually it should have been sunk within C at some point,
        # do it here to avoid a critical warning.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)
            vfuncs.object_ref()._ref_sink()
        # However, that means that the above comment of "sinks and owns" is wrong.

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)

    def test_vfunc_in_floating_transfer_full_with_held_floating(self):
        vfuncs = self.VFuncs()
        ref_count, is_floating = vfuncs.get_ref_info_for_vfunc_in_object_transfer_full(
            self.VFuncs.Object
        )
        gc.collect()

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(
            vfuncs.in_object_grefcount, 1
        )  # python wrapper sinks and owns the gobject
        self.assertFalse(vfuncs.in_object_is_floating)

        # Ref count from the perspective of C after the vfunc is called
        self.assertEqual(ref_count, 1)  # held by wrapper
        self.assertFalse(is_floating)

        # Current ref count
        self.assertEqual(vfuncs.object_ref().__grefcount__, 1)

        held_object_ref = weakref.ref(vfuncs.object_ref())
        del vfuncs.object_ref
        gc.collect()
        self.assertTrue(held_object_ref() is None)


class TestArgumentTypeErrors(unittest.TestCase):
    def test_object_argument_type_error(self):
        # ensure TypeError is raised for things which are not GObjects
        obj = Regress.TestObj()
        obj.set_bare(GObject.Object())
        obj.set_bare(None)

        self.assertRaises(TypeError, obj.set_bare, object())
        self.assertRaises(TypeError, obj.set_bare, 42)
        self.assertRaises(TypeError, obj.set_bare, "not an object")

    def test_instance_argument_error(self):
        # ensure TypeError is raised for non Regress.TestObj instances.
        obj = Regress.TestObj()
        self.assertEqual(Regress.TestObj.instance_method(obj), -1)
        self.assertRaises(TypeError, Regress.TestObj.instance_method, object())
        self.assertRaises(TypeError, Regress.TestObj.instance_method, GObject.Object())
        self.assertRaises(TypeError, Regress.TestObj.instance_method, 42)
        self.assertRaises(TypeError, Regress.TestObj.instance_method, "not an object")

    def test_instance_argument_base_type_error(self):
        # ensure TypeError is raised when a base type is passed to something
        # expecting a derived type
        obj = Regress.TestSubObj()
        self.assertEqual(Regress.TestSubObj.instance_method(obj), 0)
        self.assertRaises(
            TypeError, Regress.TestSubObj.instance_method, GObject.Object()
        )
        self.assertRaises(
            TypeError, Regress.TestSubObj.instance_method, Regress.TestObj()
        )



import os
import sys
import pytest
import unittest

try:
    if sys.platform != "win32":
        try:
            # cpython <= 3.13
            from test.test_asyncio.test_events import (
                UnixEventLoopTestsMixin as GLibEventLoopTestsMixin,
            )
        except:
            # cpython >= 3.14
            from test.test_asyncio.test_events import (
                EventLoopTestsMixin as GLibEventLoopTestsMixin,
            )
    else:
        from test.test_asyncio.test_events import EventLoopTestsMixin

        # There is Mixin for the ProactorEventLoop, so copy the skips
        class GLibEventLoopTestsMixin(EventLoopTestsMixin):
            def test_reader_callback(self):
                raise unittest.SkipTest("IocpEventLoop does not have add_reader()")

            def test_reader_callback_cancel(self):
                raise unittest.SkipTest("IocpEventLoop does not have add_reader()")

            def test_writer_callback(self):
                raise unittest.SkipTest("IocpEventLoop does not have add_writer()")

            def test_writer_callback_cancel(self):
                raise unittest.SkipTest("IocpEventLoop does not have add_writer()")

            def test_remove_fds_after_closing(self):
                raise unittest.SkipTest("IocpEventLoop does not have add_reader()")

    from test.test_asyncio.test_subprocess import SubprocessMixin
    from test.test_asyncio.utils import TestCase
except:
    assert "CI" not in os.environ, "Could not find asyncio test suite in CI!"

    class GLibEventLoopTestsMixin:
        def test_unix_event_loop_tests_missing(self):
            import warnings

            warnings.warn("No EventLoopTestsMixin found, not running tests!")
            self.skipTest("No EventLoopTestsMixin found, not running tests!")

    class SubprocessMixin:
        def test_subprocess_mixin_tests_missing(self):
            import warnings

            warnings.warn("SubprocessMixin is unavailable, not running tests!")
            self.skipTest("SubprocessMixin is unavailable, not running tests!")

    from unittest import TestCase

import gi
import gi.events
import asyncio
import signal
import socket
import threading
from gi.repository import GLib, Gio

try:
    from gi.repository import Gtk
except ImportError:
    Gtk = None


GTK4 = Gtk and Gtk._version == "4.0"


class GLibEventLoopTests(GLibEventLoopTestsMixin, TestCase):
    def __init__(self, *args):
        super().__init__(*args)
        self.loop = None

    def setUp(self):
        # Ensure a previous test did not leave an event loop around
        assert gi.events.GLibEventLoopPolicy._get_event_loop() is None, (
            "A previous test appears to have left an EventLoop open"
        )

        super().setUp()

    def _cleanup_glib_event_loop(self, loop):
        if not loop.is_closed():
            loop.close()
            raise AssertionError("Loop was no closed by the test")

    def create_event_loop(self):
        loop = gi.events.GLibEventLoop(GLib.MainContext())

        self.addCleanup(self._cleanup_glib_event_loop, loop)

        return loop


class SubprocessWatcherTests(SubprocessMixin, TestCase):
    def setUp(self):
        super().setUp()
        policy = gi.events.GLibEventLoopPolicy()
        asyncio.set_event_loop_policy(policy)
        self.loop = policy.get_event_loop()

    def tearDown(self):
        asyncio.set_event_loop_policy(None)
        self.loop.close()
        super().tearDown()

    def test_subprocess_read_pipe_cancelled(self):
        raise unittest.SkipTest(
            "GLib event loop can not be run with plain asyncio.run()"
        )

    def test_subprocess_read_write_pipe_cancelled(self):
        raise unittest.SkipTest(
            "GLib event loop can not be run with plain asyncio.run()"
        )

    def test_subprocess_write_pipe_cancelled(self):
        raise unittest.SkipTest(
            "GLib event loop can not be run with plain asyncio.run()"
        )


class GLibEventLoopPolicyTests(unittest.TestCase):
    def create_policy(self):
        return gi.events.GLibEventLoopPolicy()

    def test_get_event_loop(self):
        policy = self.create_policy()
        loop = policy.get_event_loop()
        self.assertIsInstance(loop, gi.events.GLibEventLoop)
        self.assertIs(loop, policy.get_event_loop())
        loop.close()

    def test_new_event_loop(self):
        policy = self.create_policy()
        loop = policy.new_event_loop()
        self.assertIsInstance(loop, gi.events.GLibEventLoop)
        loop.close()

        # Attaching a loop to the main thread fails
        with self.assertRaises(RuntimeError):
            policy.set_event_loop(loop)

    def test_application(self):
        task_completed = False

        async def task():
            nonlocal task_completed
            await asyncio.sleep(1)
            task_completed = True

        def activate(app):
            app.hold()
            app.create_asyncio_task(task())
            GLib.timeout_add(500, app.release)

        app = Gio.Application()
        app.connect("activate", activate)
        with self.create_policy():
            app.run()

        self.assertTrue(task_completed)

    def test_implicit_close(self):
        """Verify that implicitly closing the EventLoop (from __del__) works."""
        task_completed = False

        async def task():
            nonlocal task_completed
            await asyncio.sleep(1)
            task_completed = True

        policy = self.create_policy()
        loop = policy.new_event_loop()

        loop.run_until_complete(task())

        self.assertTrue(task_completed)

        with pytest.warns(ResourceWarning, match="unclosed event loop"):
            del loop

            # For some reason, PyPy needs two collect() steps
            import gc

            gc.collect()
            gc.collect()

    def test_nested_context_iteration(self):
        policy = self.create_policy()
        loop = policy.new_event_loop()

        called = False

        def cb():
            nonlocal called
            called = True

        async def run():
            nonlocal loop, called

            loop.call_soon(cb)
            self.assertEqual(called, False)

            # Iterating the main context does not cause cb to be called
            while loop._context.iteration(False):
                pass
            self.assertEqual(called, False)

            # Awaiting on anything *does* cause the cb to fire
            await asyncio.sleep(0)
            self.assertEqual(called, True)

        loop.run_until_complete(run())
        loop.close()

    def test_thread_event_loop(self):
        policy = self.create_policy()
        loop = policy.new_event_loop()

        res = []

        def thread_func(res):
            try:
                # We cannot get an event loop for the current thread
                with self.assertRaises(RuntimeError):
                    policy.get_event_loop()

                # We can attach our loop
                policy.set_event_loop(loop)
                # Now we can get it, and it is the same
                self.assertIs(policy.get_event_loop(), loop)

                # Simple call_soon test
                results = []

                def callback(arg1, arg2):
                    results.append((arg1, arg2))
                    loop.stop()

                loop.call_soon(callback, "hello", "world")
                loop.run_forever()
                self.assertEqual(results, [("hello", "world")])

                # We can detach it again
                policy.set_event_loop(None)

                # Which means we have none and get a runtime error
                with self.assertRaises(RuntimeError):
                    policy.get_event_loop()
            except:
                res += sys.exc_info()

        # Initially, the thread has no event loop
        thread = threading.Thread(target=lambda: thread_func(res))
        thread.start()
        thread.join()

        if res:
            t, v, tb = res
            raise t(v).with_traceback(tb)

        loop.close()

    def test_outside_context_iteration(self):
        """Iterating the main context from the outside, does not cause the
        EventLoop to dispatch.
        """
        policy = self.create_policy()
        loop = policy.new_event_loop()

        called = False

        def cb():
            nonlocal called
            called = True

        loop.call_soon(cb)
        while loop._context.iteration(False):
            pass
        loop.close()
        self.assertEqual(called, False)

    def test_inside_context_iteration(self):
        """Iterating the main context from the inside, does not cause the
        EventLoop to dispatch.
        """
        policy = self.create_policy()
        loop = policy.get_event_loop()

        done = asyncio.Future(loop=loop)

        called = False

        def cb():
            nonlocal called
            called = True

        def ctx_iterate():
            nonlocal called

            loop.call_soon(cb)
            while loop._context.iteration(False):
                pass
            self.assertEqual(called, False)

            # If we by-pass the override, then the callback is called
            while super(GLib.MainContext, loop._context).iteration(False):
                pass
            self.assertEqual(called, True)

            # It'll also be called (again) before run_until_complete finishes
            called = False
            loop.call_soon(cb)

            done.set_result(True)

            return GLib.SOURCE_REMOVE

        GLib.idle_add(ctx_iterate)
        loop.run_until_complete(done)
        loop.close()
        self.assertEqual(called, True)

    @unittest.skipUnless(Gtk, "no Gtk")
    def test_recursive_stop(self):
        """Calling stop() on the EventLoop will quit it, even if iteration
        is done recursively.
        """
        policy = self.create_policy()
        asyncio.set_event_loop_policy(policy)
        self.addCleanup(asyncio.set_event_loop_policy, None)
        loop = policy.get_event_loop()

        if not GTK4:

            def main_gtk():
                GLib.idle_add(loop.stop)
                Gtk.main()

            GLib.idle_add(main_gtk)
            Gtk.main()

        def main_glib():
            GLib.idle_add(loop.stop)
            GLib.MainLoop().run()

        GLib.idle_add(main_glib)
        GLib.MainLoop().run()

        loop.close()

    def test_glib_task_prio(self):
        """Check that we can set a task priority."""
        policy = self.create_policy()
        loop = policy.new_event_loop()

        order = []

        async def run_prio(priority):
            # Note that asyncio.sleep(0) is a special case that not sleep
            nonlocal order
            await asyncio.sleep(0)
            order.append(priority)
            await asyncio.sleep(0)
            order.append(priority)
            await asyncio.sleep(0)
            order.append(priority)

        async def run():
            t1 = asyncio.create_task(run_prio(GLib.PRIORITY_DEFAULT_IDLE))
            t1.set_priority(GLib.PRIORITY_DEFAULT_IDLE)
            t2 = asyncio.create_task(run_prio(GLib.PRIORITY_DEFAULT))
            t2.set_priority(GLib.PRIORITY_DEFAULT)
            t3 = asyncio.create_task(run_prio(GLib.PRIORITY_HIGH))
            t3.set_priority(GLib.PRIORITY_HIGH)

            pending = (t1, t2, t3)
            while pending:
                _, pending = await asyncio.wait(pending)

        loop.run_until_complete(run())
        loop.close()

        # Check that the order was correct
        self.assertEqual(
            order,
            [GLib.PRIORITY_HIGH] * 3
            + [GLib.PRIORITY_DEFAULT] * 3
            + [GLib.PRIORITY_DEFAULT_IDLE] * 3,
        )

    @unittest.skipIf(sys.platform == "win32", "add reader/writer not implemented")
    def test_source_fileobj_fd(self):
        """Regression test for
        https://gitlab.gnome.org/GNOME/pygobject/-/issues/689
        """

        class Echo:
            def __init__(self, sock, expect_bytes):
                self.sock = sock
                self.sent_bytes = 0
                self.expect_bytes = expect_bytes
                self.done = asyncio.Future()
                self.data = b""

            def send(self):
                if self.done.done():
                    return
                if self.sent_bytes < len(self.data):
                    self.sent_bytes += self.sock.send(self.data[self.sent_bytes :])
                if self.sent_bytes >= self.expect_bytes:
                    self.done.set_result(None)
                    self.sock.shutdown(socket.SHUT_WR)

            def recv(self):
                if self.done.done():
                    return
                self.data += self.sock.recv(self.expect_bytes)
                if len(self.data) >= self.expect_bytes:
                    self.sock.shutdown(socket.SHUT_RD)

        async def run():
            loop = asyncio.get_running_loop()
            s1, s2 = socket.socketpair()
            sample = b"Hello!"
            e = Echo(s1, len(sample))
            # register using file object and file descriptor
            loop.add_reader(s1, e.recv)
            loop.add_writer(s1.fileno(), e.send)
            s2.sendall(sample)
            await asyncio.wait_for(e.done, timeout=2.0)
            echo = b""
            for _ in range(len(sample)):
                echo += s2.recv(len(sample))
                if len(echo) == len(sample):
                    break
            # remove using file object and file descriptor
            loop.remove_reader(s1)
            loop.remove_writer(s1.fileno())
            s1.close()
            s2.close()
            # check if the data was echoed correctly
            self.assertEqual(sample, echo)

        policy = self.create_policy()
        loop = policy.get_event_loop()
        loop.run_until_complete(run())
        loop.close()

    @unittest.skipIf(
        sys.version_info < (3, 12),
        "Older python versions do not have the loop_factory parameter",
    )
    def test_asyncio_run(self):
        coro_state = 0

        async def coro():
            nonlocal coro_state
            coro_state = 1

            await asyncio.sleep(1)
            coro_state = 2

            # asyncio.run registered a SIGINT handler to quit
            if sys.platform == "win32":
                signal.raise_signal(signal.SIGINT)
            else:
                os.kill(os.getpid(), signal.SIGINT)

            # The signal is processed when we return to the mainloop
            await asyncio.sleep(0)
            coro_state = 3

        with self.assertRaises(KeyboardInterrupt):
            asyncio.run(coro(), loop_factory=gi.events.GLibEventLoop)

        # This works fine a second time as the first loop is unregistered (and closed)
        with self.assertRaises(KeyboardInterrupt):
            asyncio.run(coro(), loop_factory=gi.events.GLibEventLoop)

        self.assertEqual(coro_state, 2)

    def test_eventloop_context(self):
        # Main thread has only the implicit default main context
        self.assertIsNone(GLib.MainContext.get_thread_default())

        loop = gi.events.GLibEventLoop(main_context=GLib.MainContext())

        def check_loop_cannot_run(check_loop):
            with self.assertRaises(RuntimeError), check_loop:
                pass

            with self.assertRaises(RuntimeError):
                future = loop_samectx.create_future()
                future.set_result(True)
                loop_samectx.run_until_complete(future)

        def check_loop_can_run(check_loop):
            with check_loop:
                pass

            future = loop_samectx.create_future()
            future.set_result(True)
            loop_samectx.run_until_complete(future)

        with loop:
            # Entering the context manager sets the thread default
            self.assertEqual(
                hash(loop._context), hash(GLib.MainContext.get_thread_default())
            )

            loop_samectx = gi.events.GLibEventLoop()
            self.assertEqual(hash(loop._context), hash(loop_samectx._context))
            check_loop_cannot_run(loop_samectx)

            # We can do the same excercise with another main context
            loop_otherctx = gi.events.GLibEventLoop(GLib.MainContext())
            check_loop_cannot_run(loop_otherctx)

        # But, once we are outside it works fine.
        check_loop_can_run(loop_samectx)
        check_loop_can_run(loop_otherctx)



import unittest

from gi.repository import GObject
import testhelper


GUnknown = GObject.type_from_name("TestUnknown")
Unknown = GUnknown.pytype


class MyUnknown(Unknown, testhelper.Interface):
    some_property = GObject.Property(type=str)

    def __init__(self):
        Unknown.__init__(self)
        self.called = False

    def do_iface_method(self):
        self.called = True
        Unknown.do_iface_method(self)


GObject.type_register(MyUnknown)


class MyObject(GObject.GObject, testhelper.Interface):
    some_property = GObject.Property(type=str)

    def __init__(self):
        GObject.GObject.__init__(self)
        self.called = False

    def do_iface_method(self):
        self.called = True


GObject.type_register(MyObject)


class TestIfaceImpl(unittest.TestCase):
    def test_reimplement_interface(self):
        m = MyUnknown()
        m.iface_method()
        self.assertEqual(m.called, True)

    def test_implement_interface(self):
        m = MyObject()
        m.iface_method()
        self.assertEqual(m.called, True)



import sys
import os
import unittest
import warnings

import pytest

from gi.repository import GLib
from gi import PyGIDeprecationWarning

from .helper import capture_gi_deprecation_warnings


def test_child_watch_add_get_args_various():
    def cb(pid, status):
        return None

    get_args = GLib._child_watch_add_get_args
    pid = 42
    with capture_gi_deprecation_warnings():
        assert get_args(pid, cb, 2) == (0, pid, cb, (2,))

        with pytest.raises(TypeError):
            get_args(pid, cb, 2, 3, 4)

        assert get_args(0, pid, 2, 3, function=cb) == (0, pid, cb, (2, 3))
        assert get_args(0, pid, cb, 2, 3) == (0, pid, cb, (2, 3))
        assert get_args(0, pid, cb, data=99) == (0, pid, cb, (99,))

        with pytest.raises(TypeError):
            get_args(0, pid, 24)

        with pytest.raises(TypeError):
            get_args(0, pid, cb, 2, 3, data=99)


@unittest.skipIf(os.name == "nt", "not on Windows")
class TestProcess(unittest.TestCase):
    def test_deprecated_child_watch_no_data(self):
        def cb(pid, status):
            return None

        pid = object()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            res = GLib._child_watch_add_get_args(pid, cb)
            self.assertTrue(issubclass(w[0].category, PyGIDeprecationWarning))

        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], GLib.PRIORITY_DEFAULT)
        self.assertEqual(res[1], pid)
        self.assertTrue(callable(cb))
        self.assertSequenceEqual(res[3], [])

    def test_deprecated_child_watch_data_priority(self):
        def cb(pid, status):
            return None

        pid = object()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            res = GLib._child_watch_add_get_args(pid, cb, 12345, GLib.PRIORITY_HIGH)
            self.assertTrue(issubclass(w[0].category, PyGIDeprecationWarning))

        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], GLib.PRIORITY_HIGH)
        self.assertEqual(res[1], pid)
        self.assertEqual(res[2], cb)
        self.assertSequenceEqual(res[3], [12345])

    def test_deprecated_child_watch_data_priority_kwargs(self):
        def cb(pid, status):
            return None

        pid = object()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            res = GLib._child_watch_add_get_args(
                pid, cb, priority=GLib.PRIORITY_HIGH, data=12345
            )
            self.assertTrue(issubclass(w[0].category, PyGIDeprecationWarning))

        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], GLib.PRIORITY_HIGH)
        self.assertEqual(res[1], pid)
        self.assertEqual(res[2], cb)
        self.assertSequenceEqual(res[3], [12345])

    @unittest.expectedFailure  # using keyword args is fully supported by PyGObject machinery
    def test_child_watch_all_kwargs(self):
        def cb(pid, status):
            return None

        pid = object()

        res = GLib._child_watch_add_get_args(
            priority=GLib.PRIORITY_HIGH, pid=pid, function=cb, data=12345
        )
        self.assertEqual(len(res), 4)
        self.assertEqual(res[0], GLib.PRIORITY_HIGH)
        self.assertEqual(res[1], pid)
        self.assertEqual(res[2], cb)
        self.assertSequenceEqual(res[3], [12345])

    def test_child_watch_no_data(self):
        def cb(pid, status):
            self.status = status
            self.loop.quit()

        self.status = None
        self.loop = GLib.MainLoop()
        argv = [sys.executable, "-c", "import sys"]
        pid, _stdin, _stdout, _stderr = GLib.spawn_async(
            argv, flags=GLib.SpawnFlags.DO_NOT_REAP_CHILD
        )
        pid.close()
        id = GLib.child_watch_add(GLib.PRIORITY_HIGH, pid, cb)
        self.assertEqual(
            self.loop.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )
        self.loop.run()
        self.assertEqual(self.status, 0)

    def test_child_watch_with_data(self):
        def cb(pid, status, data):
            self.status = status
            self.data = data
            self.loop.quit()

        self.data = None
        self.status = None
        self.loop = GLib.MainLoop()
        argv = [sys.executable, "-c", "import sys"]
        pid, stdin, stdout, stderr = GLib.spawn_async(
            argv, flags=GLib.SpawnFlags.DO_NOT_REAP_CHILD
        )
        self.assertEqual(stdin, None)
        self.assertEqual(stdout, None)
        self.assertEqual(stderr, None)
        self.assertNotEqual(pid, 0)
        pid.close()
        id = GLib.child_watch_add(GLib.PRIORITY_HIGH, pid, cb, 12345)
        self.assertEqual(
            self.loop.get_context().find_source_by_id(id).priority, GLib.PRIORITY_HIGH
        )
        self.loop.run()
        self.assertEqual(self.data, 12345)
        self.assertEqual(self.status, 0)

    def test_spawn_async_fds(self):
        pid, stdin, stdout, stderr = GLib.spawn_async(
            ["cat"],
            flags=GLib.SpawnFlags.SEARCH_PATH,
            standard_input=True,
            standard_output=True,
            standard_error=True,
        )
        os.write(stdin, b"hello world!\n")
        os.close(stdin)
        out = os.read(stdout, 50)
        os.close(stdout)
        err = os.read(stderr, 50)
        os.close(stderr)
        pid.close()
        self.assertEqual(out, b"hello world!\n")
        self.assertEqual(err, b"")

    def test_spawn_async_with_pipes(self):
        _res, pid, stdin, stdout, stderr = GLib.spawn_async_with_pipes(
            working_directory=None,
            argv=["cat"],
            envp=None,
            flags=GLib.SpawnFlags.SEARCH_PATH,
        )

        os.write(stdin, b"hello world!\n")
        os.close(stdin)
        out = os.read(stdout, 50)
        os.close(stdout)
        err = os.read(stderr, 50)
        os.close(stderr)
        GLib.spawn_close_pid(pid)
        self.assertEqual(out, b"hello world!\n")
        self.assertEqual(err, b"")

    def test_spawn_async_envp(self):
        pid, stdin, stdout, stderr = GLib.spawn_async(
            ["sh", "-c", "echo $TEST_VAR"],
            ["TEST_VAR=moo!"],
            flags=GLib.SpawnFlags.SEARCH_PATH,
            standard_output=True,
        )
        self.assertEqual(stdin, None)
        self.assertEqual(stderr, None)
        out = os.read(stdout, 50)
        os.close(stdout)
        pid.close()
        self.assertEqual(out, b"moo!\n")

    def test_backwards_compat_flags(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", PyGIDeprecationWarning)

            self.assertEqual(
                GLib.SpawnFlags.DO_NOT_REAP_CHILD, GLib.SPAWN_DO_NOT_REAP_CHILD
            )



# Copyright (C) 2015 Christoph Reiter <reiter.christoph@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import unittest
import pickle

import gi
from gi.repository import GIMarshallingTests
from gi.repository import Regress


ResultTuple = gi._gi.ResultTuple


class TestResultTuple(unittest.TestCase):
    def test_base(self):
        self.assertTrue(issubclass(ResultTuple, tuple))

    def test_create(self):
        names = [None, "foo", None, "bar"]
        for i in range(10):
            new = ResultTuple._new_type(names)
            self.assertTrue(issubclass(new, ResultTuple))

    def test_repr_dir(self):
        new = ResultTuple._new_type([None, "foo", None, "bar"])
        inst = new([1, 2, 3, "a"])

        self.assertEqual(repr(inst), "(1, foo=2, 3, bar='a')")
        self.assertTrue("foo" in dir(inst))

    def test_repr_dir_empty(self):
        new = ResultTuple._new_type([])
        inst = new()
        self.assertEqual(repr(inst), "()")
        dir(inst)

    def test_getatttr(self):
        new = ResultTuple._new_type([None, "foo", None, "bar"])
        inst = new([1, 2, 3, "a"])

        self.assertTrue(hasattr(inst, "foo"))
        self.assertEqual(inst.foo, inst[1])
        self.assertRaises(AttributeError, getattr, inst, "nope")

    def test_pickle(self):
        new = ResultTuple._new_type([None, "foo", None, "bar"])
        inst = new([1, 2, 3, "a"])

        inst2 = pickle.loads(pickle.dumps(inst))
        self.assertEqual(inst2, inst)
        self.assertTrue(isinstance(inst2, tuple))
        self.assertFalse(isinstance(inst2, new))

    def test_gi(self):
        res = GIMarshallingTests.init_function([])
        self.assertEqual(repr(res), "(True, argv=[])")

        res = GIMarshallingTests.array_return_etc(5, 9)
        self.assertEqual(repr(res), "([5, 0, 1, 9], sum=14)")

        res = GIMarshallingTests.array_out_etc(-5, 9)
        self.assertEqual(repr(res), "(ints=[-5, 0, 1, 9], sum=4)")

        def cb():
            return 1, 2

        res = GIMarshallingTests.callback_multiple_out_parameters(cb)
        self.assertEqual(repr(res), "(a=1.0, b=2.0)")

    def test_regress(self):
        res = Regress.TestObj().skip_return_val(50, 42.0, 60, 2, 3)
        self.assertEqual(repr(res), "(out_b=51, inout_d=61, out_sum=32)")
